from api_fhir_r4.configurations import R4SubscriptionConfig
from api_fhir_r4.models import Subscription


class SubscriptionStatusMapping:
    _IMIS_OFF = Subscription.SubscriptionStatus.INACTIVE.value
    _FHIR_OFF = R4SubscriptionConfig.get_fhir_subscription_status_off()
    _IMIS_ACTIVE = Subscription.SubscriptionStatus.ACTIVE.value
    _FHIR_ACTIVE = R4SubscriptionConfig.get_fhir_subscription_status_active()

    to_fhir_status = {_IMIS_OFF: _FHIR_OFF, _IMIS_ACTIVE: _FHIR_ACTIVE}

    to_imis_status = {_FHIR_OFF: _IMIS_OFF, _FHIR_ACTIVE: _IMIS_ACTIVE}


class SubscriptionChannelMapping:
    _IMIS_REST_HOOK = Subscription.SubscriptionChannel.REST_HOOK.value
    _FHIR_REST_HOOK = R4SubscriptionConfig.get_fhir_subscription_channel_rest_hook()

    to_fhir_channel = {_IMIS_REST_HOOK: _FHIR_REST_HOOK}

    to_imis_channel = {_FHIR_REST_HOOK: _IMIS_REST_HOOK}



from api_fhir_r4.converters import PolicyHolderOrganisationConverter
from api_fhir_r4.tests.mixin import ConvertToFhirTestMixin
from api_fhir_r4.tests.mixin.policyHolderOrganisationTestMixin import (
    PolicyHolderOrganisationTestMixin,
)


class PolicyHolderOrganisationConverterTestCase(
    PolicyHolderOrganisationTestMixin, ConvertToFhirTestMixin
):
    converter = PolicyHolderOrganisationConverter



from api_fhir_r4.configurations import CoverageEligibilityConfiguration


class R4CoverageEligibilityConfiguration(CoverageEligibilityConfiguration):

    @classmethod
    def build_configuration(cls, cfg):
        cls.get_config().R4_fhir_coverage_eligibility_config = cfg[
            "R4_fhir_coverage_eligibility_config"
        ]

    @classmethod
    def get_serializer(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_serializer"
        )

    @classmethod
    def get_fhir_financial_code(cls):
        pass

    @classmethod
    def get_fhir_item_code(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_item_code", "item"
        )

    @classmethod
    def get_fhir_service_code(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_service_code", "service"
        )

    @classmethod
    def get_fhir_total_admissions_code(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_total_admissions_code", "total_admissions"
        )

    @classmethod
    def get_fhir_total_visits_code(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_total_visits_code", "total_visits"
        )

    @classmethod
    def get_fhir_total_consultations_code(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_total_consultations_code", "total_consultations"
        )

    @classmethod
    def get_fhir_total_surgeries_code(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_total_surgeries_code", "total_surgeries"
        )

    @classmethod
    def get_fhir_total_deliveries_code(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_total_deliveries_code", "total_deliveries"
        )

    @classmethod
    def get_fhir_total_antenatal_code(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_total_antenatal_code", "total_antenatal"
        )

    @classmethod
    def get_fhir_consultation_amount_code(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_consultation_amount_code", "consultation_amount"
        )

    @classmethod
    def get_fhir_surgery_amount_code(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_surgery_amount_code", "surgery_amount"
        )

    @classmethod
    def get_fhir_delivery_amount_code(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_delivery_amount_code", "delivery_amount"
        )

    @classmethod
    def get_fhir_hospitalization_amount_code(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_hospitalization_amount_code", "hospitalization_amount"
        )

    @classmethod
    def get_fhir_antenatal_amount_code(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_antenatal_amount_code", "antenatal_amount"
        )

    @classmethod
    def get_fhir_service_left_code(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_service_left_code", "service_left"
        )

    @classmethod
    def get_fhir_item_left_code(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_item_left_code", "item_left"
        )

    @classmethod
    def get_fhir_is_item_ok_code(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_is_item_ok_code", "is_item_ok"
        )

    @classmethod
    def get_fhir_is_service_ok_code(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_is_service_ok_code", "is_service_ok"
        )

    @classmethod
    def get_fhir_balance_code(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_balance_code", "balance"
        )

    @classmethod
    def get_fhir_balance_default_category(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_balance_default_category", "medical"
        )

    @classmethod
    def get_fhir_active_policy_status(cls):
        return cls.get_config_attribute("R4_fhir_coverage_eligibility_config").get(
            "fhir_active_policy_status", ("A", 2)
        )



import copy

from product.models import Product
from api_fhir_r4.converters import InsurancePlanConverter
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.serializers.baseSerializer import BaseFHIRSerializer


class InsurancePlanSerializer(BaseFHIRSerializer):
    fhirConverter = InsurancePlanConverter

    def create(self, validated_data):
        code = validated_data.get("code")
        if Product.objects.filter(code=code).count() > 0:
            raise FHIRException("Exists product with following code `{}`".format(code))
        copied_data = copy.deepcopy(validated_data)
        if "_state" in copied_data:
            del copied_data["_state"]
        # TODO services in product hasn't been developed yet.
        return Product.objects.create(**copied_data)

    def update(self, instance, validated_data):
        instance.code = validated_data.get("code", instance.code)
        instance.name = validated_data.get("name", instance.name)
        instance.date_from = validated_data.get("date_from", instance.date_from)
        instance.date_to = validated_data.get("date_to", instance.date_from)
        instance.location = validated_data.get("location", instance.location)
        instance.member_count = validated_data.get(
            "member_count", instance.member_count
        )
        instance.insurance_period = validated_data.get(
            "insurance_period", instance.insurance_period
        )
        instance.lump_sum = validated_data.get("lump_sum", instance.lump_sum)
        instance.threshold = validated_data.get("threshold", instance.threshold)
        instance.premium_adult = validated_data.get(
            "premium_adult", instance.premium_adult
        )
        instance.premium_child = validated_data.get(
            "premium_child", instance.premium_child
        )
        instance.registration_lump_sum = validated_data.get(
            "registration_lump_sum", instance.registration_lump_sum
        )
        instance.registration_fee = validated_data.get(
            "registration_fee", instance.registration_fee
        )
        instance.general_assembly_lump_sum = validated_data.get(
            "general_assembly_lump_sum", instance.general_assembly_lump_sum
        )
        instance.general_assembly_fee = validated_data.get(
            "general_assembly_fee", instance.general_assembly_fee
        )
        instance.conversion_product = validated_data.get(
            "conversion_product", instance.conversion_product
        )
        instance.max_installments = validated_data.get(
            "max_installments", instance.max_installments
        )
        instance.grace_period_enrolment = validated_data.get(
            "grace_period_enrolment", instance.grace_period_enrolment
        )
        instance.audit_user_id = self.get_audit_user_id()
        instance.save()
        return instance



from api_fhir_r4.tests import GenericTestMixin
from invoice.models import Invoice, PaymentInvoice, DetailPaymentInvoice


class PaymentNoticeTestMixin(GenericTestMixin):
    _TEST_PAYMENT_NOTICE_IMIS_RECONCILIATION_STATUS = (
        PaymentInvoice.ReconciliationStatus.RECONCILIATED
    )
    _TEST_PAYMENT_NOTICE_IMIS_JSON_EXT = {
        "reconciliation": {"id": "id-renconiliation-test-1"}
    }
    _TEST_PAYMENT_NOTICE_IMIS_DATE_PAYMENT = "2022-05-17"
    _TEST_PAYMENT_NOTICE_IMIS_DATE_CREATED = "2022-05-17T10:58:46.807554"
    _TEST_PAYMENT_NOTICE_IMIS_AMOUNT_RECEIVED = 10000.0
    _TEST_PAYMENT_NOTICE_IMIS_DETAILS_STATUS = (
        DetailPaymentInvoice.DetailPaymentStatus.ACCEPTED
    )
    _TEST_PAYMENT_NOTICE_IMIS_DETAILS_PAYMENT = None
    _TEST_PAYMENT_NOTICE_IMIS_DETAILS_SUBJECT_ID = (
        "bd84b2f2-ec1d-48de-8f8c-5a477aa4a29f"
    )
    _TEST_PAYMENT_NOTICE_IMIS_DETAILS_RECON_ID = "id-renconiliation-test-1"

    _TEST_PAYMENT_NOTICE_FHIR_STATUS = "active"
    _TEST_PAYMENT_NOTICE_FHIR_REQUEST_REFERENCE = (
        "Invoice/bd84b2f2-ec1d-48de-8f8c-5a477aa4a29f"
    )
    _TEST_PAYMENT_NOTICE_FHIR_REQUEST_REFERENCE_TYPE = "Invoice"
    _TEST_PAYMENT_NOTICE_FHIR_CREATED = "2022-05-17T10:58:46.807554"
    _TEST_PAYMENT_NOTICE_FHIR_DATE_PAYMENT = "2022-05-17"
    _TEST_PAYMENT_NOTICE_FHIR_PAYMENT_REFERENCE = (
        "PaymentReconciliation/id-renconiliation-test-1"
    )
    _TEST_PAYMENT_NOTICE_FHIR_RECIPIENT = "Organization/openIMIS-Implementation"

    _TEST_PAYMENT_NOTICE_FHIR_AMOUNT_VALUE = "10000.0"
    _TEST_PAYMENT_NOTICE_FHIR_AMOUNT_CURRENCY = "$"

    _TEST_PAYMENT_NOTICE_FHIR_PAYMENT_STATUS = "cleared"

    def create_test_imis_instance(self):
        imis_payment = PaymentInvoice(
            **{
                "reconciliation_status": self._TEST_PAYMENT_NOTICE_IMIS_RECONCILIATION_STATUS,
                "amount_received": self._TEST_PAYMENT_NOTICE_IMIS_AMOUNT_RECEIVED,
                "date_payment": self._TEST_PAYMENT_NOTICE_IMIS_DATE_PAYMENT,
                "json_ext": self._TEST_PAYMENT_NOTICE_IMIS_JSON_EXT,
                "payer_ref": "ICD-54546",
            }
        )
        self._TEST_PAYMENT_NOTICE_IMIS_DETAILS_PAYMENT = imis_payment
        imis_payment_detail = DetailPaymentInvoice(
            **{
                "status": self._TEST_PAYMENT_NOTICE_IMIS_DETAILS_STATUS,
                "amount": self._TEST_PAYMENT_NOTICE_IMIS_AMOUNT_RECEIVED,
                "payment": self._TEST_PAYMENT_NOTICE_IMIS_DETAILS_PAYMENT,
                "subject_id": self._TEST_PAYMENT_NOTICE_IMIS_DETAILS_SUBJECT_ID,
            }
        )
        return imis_payment, imis_payment_detail

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(
            imis_obj.reconciliation_status,
            self._TEST_PAYMENT_NOTICE_IMIS_RECONCILIATION_STATUS,
        )
        self.assertEqual(
            round(float(imis_obj.amount_received), 2),
            round(float(self._TEST_PAYMENT_NOTICE_IMIS_AMOUNT_RECEIVED), 2),
        )
        self.assertEqual(
            f"{imis_obj.date_payment}", self._TEST_PAYMENT_NOTICE_IMIS_DATE_PAYMENT
        )
        self.assertEqual(
            imis_obj.json_ext["reconciliation"]["id"],
            self._TEST_PAYMENT_NOTICE_IMIS_JSON_EXT["reconciliation"]["id"],
        )
        self.assertEqual(
            imis_obj.payer_ref, "PaymentReconciliation/id-renconiliation-test-1"
        )

    def verify_imis_detail_instance(self, imis_obj, imis_detail_obj):
        self.assertEqual(
            imis_detail_obj.status, self._TEST_PAYMENT_NOTICE_IMIS_DETAILS_STATUS
        )
        self.assertEqual(
            imis_detail_obj.amount, self._TEST_PAYMENT_NOTICE_IMIS_AMOUNT_RECEIVED
        )
        self.assertEqual(imis_detail_obj.payment.id, imis_obj.id)
        self.assertEqual(
            imis_detail_obj.subject_id,
            self._TEST_PAYMENT_NOTICE_IMIS_DETAILS_SUBJECT_ID,
        )
        self.assertEqual(
            imis_detail_obj.reconcilation_id,
            self._TEST_PAYMENT_NOTICE_IMIS_DETAILS_RECON_ID,
        )

    def verify_imis_invoice_status(self, imis_invoice_status):
        self.assertEqual(imis_invoice_status, Invoice.Status.PAID)

    def create_test_fhir_instance(self):
        return {
            "status": self._TEST_PAYMENT_NOTICE_FHIR_STATUS,
            "request": {
                "reference": self._TEST_PAYMENT_NOTICE_FHIR_PAYMENT_REFERENCE,
                "type": self._TEST_PAYMENT_NOTICE_FHIR_REQUEST_REFERENCE_TYPE,
            },
            "created": self._TEST_PAYMENT_NOTICE_FHIR_CREATED,
            "payment": {"reference": self._TEST_PAYMENT_NOTICE_FHIR_PAYMENT_REFERENCE},
            "paymentDate": self._TEST_PAYMENT_NOTICE_IMIS_DATE_PAYMENT,
            "recipient": {"reference": self._TEST_PAYMENT_NOTICE_FHIR_RECIPIENT},
            "amount": {
                "value": self._TEST_PAYMENT_NOTICE_FHIR_AMOUNT_VALUE,
                "currency": "$",
            },
            "paymentStatus": {
                "coding": [
                    {
                        "system": "http://terminology.hl7.org/CodeSystem/paymentstatus",
                        "code": self._TEST_PAYMENT_NOTICE_FHIR_PAYMENT_STATUS,
                        "display": self._TEST_PAYMENT_NOTICE_FHIR_PAYMENT_STATUS,
                    }
                ]
            },
        }

    def verify_fhir_instance(self, fhir_obj):
        self.assertEqual(fhir_obj.status, self._TEST_PAYMENT_NOTICE_FHIR_STATUS)
        self.assertEqual(
            fhir_obj.request.reference, self._TEST_PAYMENT_NOTICE_FHIR_REQUEST_REFERENCE
        )
        self.assertEqual(
            f"{fhir_obj.paymentDate}", self._TEST_PAYMENT_NOTICE_FHIR_DATE_PAYMENT
        )
        self.assertEqual(
            fhir_obj.recipient.reference, self._TEST_PAYMENT_NOTICE_FHIR_RECIPIENT
        )
        self.assertEqual(
            round(float(fhir_obj.amount.value), 2),
            round(float(self._TEST_PAYMENT_NOTICE_FHIR_AMOUNT_VALUE), 2),
        )
        self.assertEqual(
            fhir_obj.paymentStatus.coding[0].code,
            self._TEST_PAYMENT_NOTICE_FHIR_PAYMENT_STATUS,
        )



from drf_spectacular.authentication import SessionScheme, TokenScheme
from drf_spectacular.plumbing import build_bearer_security_scheme_object
from drf_spectacular.utils import inline_serializer
from rest_framework import fields


class CsrfExemptAuthenticationScheme(SessionScheme):
    target_class = "api_fhir_r4.views.CsrfExemptSessionAuthentication"
    name = "Session Authentication (CSRF Exempt)"
    priority = 0
    exclude = True


class JWTAuthenticationScheme(TokenScheme):
    target_class = "core.jwt_authentication.JWTAuthentication"
    name = "JWT Authentication"
    priority = 0

    def get_security_definition(self, auto_schema):
        return build_bearer_security_scheme_object(
            header_name="Authorization",
            token_prefix="Bearer",
        )


def get_inline_error_serializer():
    return inline_serializer(
        name="InlineErrorSerializer",
        fields={
            "resourceType": fields.CharField(),
            "issue": fields.ListField(child=fields.JSONField()),
        },
    )


def get_inline_login_request_serializer():
    return inline_serializer(
        name="InlineLoginRequestSerializer",
        fields={
            "username": fields.CharField(),
            "password": fields.CharField(),
        },
    )


def get_inline_login_200_response_serializer():
    return inline_serializer(
        name="InlineLoginResponseSerializer",
        fields={
            "token": fields.CharField(),
            "exp": fields.IntegerField(),
        },
    )




from django.db.models import Model
from rest_framework.test import APITestCase
from rest_framework import status
from typing import Type

from api_fhir_r4.tests import GenericFhirAPITestMixin
from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.utils import TimeUtils
from claim.models import Claim
from core.models.user import ClaimAdmin
from location.models import HealthFacility, UserDistrict
from medical.models import Item, Service
from location.test_helpers import (
    create_test_location,
)
from api_fhir_r4.tests.utils import (
    load_and_replace_json,
    get_connection_payload,
    get_or_create_user_api,
)
from claim.test_helpers import create_test_claim_context
from insuree.models import Insuree, Family
from datetime import datetime
from core.test_helpers import create_admin_role


class ClaimAPIContainedTestBaseMixin:
    base_url = GeneralConfiguration.get_base_url() + "Claim/"
    _test_json_request_path = "/test/test_claim_contained.json"

    # diagnosis data
    _TEST_MAIN_ICD_CODE = "T_CD"
    _TEST_MAIN_ICD_NAME = "Test diagnosis"

    _TEST_CLAIM_ADMIN_UUID = "044c33d1-dbf3-4d6a-9924-3797b461e535"
    _TEST_INSUREE_UUID = "AAAA08B5-6E85-470C-83EC-0EE9370F0000"
    _TEST_INSUREE_CODE = "105000003"
    _TEST_GROUP_UUID = "AAAA5232-9054-4D86-B4F2-0E9C4ADF0000"

    # claim item data
    _TEST_ITEM_CODE = "iCode"
    _TEST_ITEM_UUID = "e2bc1546-390b-4d41-8571-632ecf7a0936"
    _TEST_ITEM_QUANTITY_PROVIDED = 10.0
    _TEST_ITEM_PRICE_ASKED = 10.0
    _TEST_ITEM_EXPLANATION = "item_explanation"
    _TEST_ITEM_TYPE = "D"

    # claim service data
    _TEST_SERVICE_CODE = "sCode"
    _TEST_SERVICE_UUID = "a17602f4-e9ff-4f42-a6a4-ccefcb23b4d6"
    _TEST_SERVICE_QUANTITY_PROVIDED = 1
    _TEST_SERVICE_PRICE_ASKED = 21000.0
    _TEST_SERVICE_EXPLANATION = "service_explanation"
    _TEST_SERVICE_TYPE = "D"

    # hf test data
    _TEST_HF_ID = 10000
    _TEST_HF_UUID = "AAAA5F9B-97C6-444B-AAD9-2FCCFD460000"
    _TEST_HF_CODE = "12345678"
    _TEST_HF_NAME = "TEST_NAME"
    _TEST_HF_LEVEL = "H"
    _TEST_HF_LEGAL_FORM = "G"
    _TEST_ADDRESS = "TEST_ADDRESS"
    _TEST_PHONE = "133-996-476"
    _TEST_FAX = "1-408-999 8888"
    _TEST_EMAIL = "TEST@TEST.com"

    _ADMIN_AUDIT_USER_ID = -1

    _test_json_path_credentials = "/test/test_login.json"
    _TEST_USER_NAME = "Admin"
    _TEST_USER_PASSWORD = "admin123"
    _TEST_DATA_USER = {
        "username": _TEST_USER_NAME,
        "last_name": _TEST_USER_NAME,
        "password": _TEST_USER_PASSWORD,
        "other_names": _TEST_USER_NAME,
        "user_types": "INTERACTIVE",
        "language": "en",
        "roles": [create_admin_role().id],
    }
    _test_request_data_credentials = None
    test_village = None
    test_hf = None
    test_claim_admin = None
    test_insuree = None
    sub_str = {}

    def setUp(self):
        super(ClaimAPIContainedTestBaseMixin, self).setUp()
        # Initialize basic test data
        self.create_dependencies()
        self.sub_str[self._TEST_GROUP_UUID] = self.test_insuree.family.uuid
        self.sub_str[self._TEST_INSUREE_UUID] = self.test_insuree.uuid
        self.sub_str[self._TEST_INSUREE_CODE] = self.test_insuree.chf_id
        self.sub_str[self._TEST_CLAIM_ADMIN_UUID] = self.test_claim_admin.uuid
        self.sub_str[self._TEST_HF_UUID] = self.test_hf.uuid
        self.sub_str[self._TEST_HF_CODE] = self.test_hf.code
        self.sub_str["2021-02-03"] = datetime.now().strftime("%Y-%m-%d")

        self._test_request_data = load_and_replace_json(
            self._test_json_request_path, self.sub_str
        )

    def create_dependencies(self):

        region = create_test_location(
            "R", custom_props={"code": "R2", "name": "Tahida"}
        )
        district = create_test_location(
            "D", custom_props={"code": "R2D2", "name": "Vida", "parent": region}
        )
        ward = create_test_location(
            "M", custom_props={"code": "R2D2M1", "name": "Majhi", "parent": district}
        )
        create_test_location(
            "V", custom_props={"code": "R2D2M1V1", "name": "Radho", "parent": ward}
        )

        self.test_claim, self.test_insuree, policy, self.test_hf = (
            create_test_claim_context(
                claim={
                    "icd": {
                        "code": self._TEST_MAIN_ICD_CODE,
                        "name": self._TEST_MAIN_ICD_NAME,
                    }
                },
                claim_admin={
                    "last_name": self._TEST_DATA_USER["last_name"],
                    "other_names": self._TEST_DATA_USER["other_names"],
                },
                insuree={},
                product={},
                hf={
                    "name": self._TEST_HF_NAME,
                    "level": self._TEST_HF_LEVEL,
                    "legal_form_id": self._TEST_HF_LEGAL_FORM,
                    "address": self._TEST_ADDRESS,
                    "phone": self._TEST_PHONE,
                    "fax": self._TEST_FAX,
                    "email": self._TEST_EMAIL,
                },
                items=[{"code": self._TEST_ITEM_CODE}],
                services=[{"code": self._TEST_SERVICE_CODE}],
            )
        )

        self._TEST_USER = get_or_create_user_api(self._TEST_DATA_USER)

        UserDistrict.objects.create(
            **{
                "user": self._TEST_USER._u,
                "location": self.test_insuree.family.location.parent.parent,
                "audit_user_id": -1,
            }
        )
        UserDistrict.objects.create(
            **{"user": self._TEST_USER._u, "location": district, "audit_user_id": -1}
        )

        self.test_claim_admin = self.test_claim.admin

        self.test_village = (
            self.test_insuree.current_village or self.test_insuree.family.location
        )
        self._TEST_HF_ID = self.test_hf.id

        ud = UserDistrict()
        ud.location = self.test_village.parent.parent
        ud.audit_user_id = self._ADMIN_AUDIT_USER_ID
        ud.user = self._TEST_USER.i_user
        ud.validity_from = TimeUtils.now()
        ud.save()

        self._TEST_USER.claim_admin = self.test_claim.admin
        self._TEST_USER.save()

    def assert_response(self, response_json):
        self.assertEqual(response_json["outcome"], "complete")
        for item in response_json["item"]:
            for adjudication in item["adjudication"]:
                self.assertEqual(
                    adjudication["category"]["coding"][0]["code"],
                    f"{Claim.STATUS_REJECTED}",
                )
                # 2 not in price list
                # FIXME familly in some cases self.assertEqual(adjudication["reason"]["coding"][0]["code"], '2')

        self.assertEqual(response_json["resourceType"], "ClaimResponse")

    def assert_hf_created(self):
        self._assert_unique_created(self.test_hf.uuid, HealthFacility)

    def assert_insuree_created(self):
        insuree = self._assert_unique_created(str(self.test_insuree.uuid), Insuree)

        family = self._assert_unique_created(str(self.test_insuree.family.uuid), Family)

        self.assertEqual(insuree, family.head_insuree)

    def assert_claim_admin_created(self):
        return None  # FIXME
        expected_claim_admin_uuid = "AAAA5229-DD11-4383-863C-E2FAD1B20000"
        self._assert_unique_created(expected_claim_admin_uuid, ClaimAdmin)
        # HF added using practitioner role
        admin = ClaimAdmin.objects.get(uuid=expected_claim_admin_uuid)
        hf = admin.health_facility
        self.assertEqual(hf.uuid, "AAAA5F9B-97C6-444B-AAD9-2FCCFD460000")

    def assert_items_created(self):
        expected_item_uuid = "AAAA76E2-DC28-4B48-8E29-3AC4ABEC0000"
        self._assert_unique_created(expected_item_uuid, Item)

        expected_service_uuid = "AAAA29BA-3F4E-4E6F-B55C-23A488A10000"
        self._assert_unique_created(expected_service_uuid, Service)

    def _assert_unique_created(self, expected_uuid, django_model: Type[Model]):
        msg = (
            f"Contained resource should create unique object of type "
            f"{django_model} with uuid {expected_uuid}"
        )

        query = django_model.objects.filter(uuid=expected_uuid).all()

        self.assertEqual(query.count(), 1, msg)
        return query.get()

    def assert_hf_updated(self):
        return None  # FIXME
        expected_updated_address = "Uitly road 1"
        hf = HealthFacility.objects.get(uuid=self.test_hf.uuid)
        self.assertEqual(hf.address, expected_updated_address)

    def assert_contained(self, json_response):
        contained = json_response["contained"]
        self.assertResourceExists(
            contained, "Patient", "AAAA08B5-6E85-470C-83EC-0EE9370F0000"
        )
        self.assertResourceExists(
            contained, "Organization", "AAAA5F9B-97C6-444B-AAD9-2FCCFD460000"
        )
        self.assertResourceExists(
            contained, "Group", "AAAA5232-9054-4D86-B4F2-0E9C4ADF0000"
        )
        self.assertResourceExists(
            contained, "Practitioner", "AAAA5229-DD11-4383-863C-E2FAD1B20000"
        )
        self.assertResourceExists(
            contained, "Medication", "AAAA76E2-DC28-4B48-8E29-3AC4ABEC0000"
        )
        self.assertResourceExists(
            contained, "ActivityDefinition", "AAAA29BA-3F4E-4E6F-B55C-23A488A10000"
        )

    def assertResourceExists(self, contained, resource_type, resource_id):
        x = [
            x
            for x in contained
            if x["resourceType"] == resource_type and x["id"] == resource_id
        ]
        self.assertEqual(len(x), 1)


class ClaimAPIContainedTests(
    ClaimAPIContainedTestBaseMixin, GenericFhirAPITestMixin, APITestCase
):
    base_url = GeneralConfiguration.get_base_url() + "Claim/"
    _test_json_path = "/test/test_claim_contained.json"

    def test_post_should_create_correctly(self):
        response = self.client.post(
            GeneralConfiguration.get_base_url() + "login/",
            data=get_connection_payload(self._TEST_DATA_USER),
            format="json",
        )
        response_json = response.json()
        token = response_json["token"]
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        headers = {
            "Content-Type": "application/json",
            "HTTP_AUTHORIZATION": f"Bearer {token}",
        }
        response = self.client.post(
            self.base_url, data=self._test_request_data, format="json", **headers
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertIsNotNone(response.content)
        response_json = response.json()
        self.assert_response(response_json)
        self.assert_hf_created()
        self.assert_insuree_created()
        self.assert_claim_admin_created()

    def test_post_should_update_contained_correctly(self):
        response = self.client.post(
            GeneralConfiguration.get_base_url() + "login/",
            data=get_connection_payload(self._TEST_DATA_USER),
            format="json",
        )
        response_json = response.json()
        token = response_json["token"]
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        headers = {
            "Content-Type": "application/json",
            "HTTP_AUTHORIZATION": f"Bearer {token}",
        }

        # Confirm HF already exists
        self.assertTrue(HealthFacility.objects.filter(uuid=self.test_hf.uuid).exists())
        response = self.client.post(
            self.base_url, data=self._test_request_data, format="json", **headers
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertIsNotNone(response.content)
        response_json = response.json()
        self.assert_response(response_json)
        self.assert_hf_created()
        self.assert_insuree_created()
        self.assert_claim_admin_created()
        # self.assert_items_created()
        self.assert_hf_updated()

    def test_get_should_return_200_claim_with_contained(self):
        # Test if get Claim return contained resources
        response = self.client.post(
            GeneralConfiguration.get_base_url() + "login/",
            data=get_connection_payload(self._TEST_DATA_USER),
            format="json",
        )
        response_json = response.json()
        token = response_json["token"]
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        headers = {
            "Content-Type": "application/json",
            "HTTP_AUTHORIZATION": f"Bearer {token}",
        }
        # Create claim
        response = self.client.post(
            self.base_url, data=self._test_request_data, format="json", **headers
        )
        response_json = response.json()
        self.assertTrue(status.is_success(response.status_code))
        uuid = response_json["id"]
        url = f"{GeneralConfiguration.get_base_url()}Claim/{uuid}/?contained=True"
        response = self.client.get(url, data=None, format="json", **headers)
        # FIXME replace fix UUID with codes self.assert_contained(response.json())
        self.assertEqual(response.status_code, status.HTTP_200_OK)



from copy import deepcopy

from drf_spectacular.utils import inline_serializer
from rest_framework import fields
from rest_framework.exceptions import ValidationError, APIException, PermissionDenied

from api_fhir_r4.configurations import R4SubscriptionConfig
from api_fhir_r4.subscriptions import SubscriptionConverter
from api_fhir_r4.models import Subscription
from api_fhir_r4.permissions import (
    FHIRApiInsureePermissions,
    FHIRApiInvoicePermissions,
    FHIRApiHealthServicePermissions,
    FHIRApiClaimPermissions,
)
from api_fhir_r4.serializers import BaseFHIRSerializer
from api_fhir_r4.services import SubscriptionService
from api_fhir_r4.mixins import RetrieveModelMixin


class SubscriptionSerializer(BaseFHIRSerializer, RetrieveModelMixin):
    fhirConverter = SubscriptionConverter
    _error_while_saving = "Error while saving a subscription: %(msg)s"

    resource_permissions = {
        "patient": FHIRApiInsureePermissions.permissions_get,
        "invoice": FHIRApiInvoicePermissions.permissions_get,
        "organisation": FHIRApiHealthServicePermissions.permissions_get,
        "claim": FHIRApiClaimPermissions.permissions_get,
    }

    def create(self, validated_data):
        user = self.user or self.context["request"].user
        self.check_resource_rights(user, validated_data)
        service = SubscriptionService(user)
        copied_data = deepcopy(validated_data)
        if "_state" in copied_data:
            del copied_data["_state"]
        if "_original_state" in copied_data:
            del copied_data["_original_state"]
        result = service.create(copied_data)
        return self.get_result_object(result)

    def update(self, instance, validated_data):
        user = self.user or self.context["request"].user
        self.check_instance_id(instance, validated_data)
        self.check_object_owner(user, instance)
        self.check_resource_rights(user, validated_data)
        service = SubscriptionService(user)
        copied_data = {
            key: value
            for key, value in deepcopy(validated_data).items()
            if value is not None
        }
        copied_data["id"] = instance.id
        if "_state" in copied_data:
            del copied_data["_state"]
        if "_original_state" in copied_data:
            del copied_data["_original_state"]
        result = service.update(copied_data)
        return self.get_result_object(result)

    def get_result_object(self, result):
        if result.get("success", False):
            return Subscription.objects.get(id=result["data"]["id"])
        else:
            raise APIException(
                self._error_while_saving % {"msg": result.get("message", "Unknown")}
            )

    def check_resource_rights(self, user, data):
        resource = (
            data.get("criteria", {})
            .get(R4SubscriptionConfig.get_fhir_sub_criteria_key_resource(), "")
            .lower()
        )
        if not resource or resource not in self.resource_permissions:
            raise ValidationError(
                self._error_while_saving % {"msg": f"Invalid resource ({resource})"}
            )

        if not user.has_perms(self.resource_permissions[resource]):
            raise PermissionDenied(
                detail=self._error_while_saving
                % {"msg": f"You have no permissions to subscribe to {resource}"}
            )

    def check_object_owner(self, user, instance):
        if str(user.id).lower() != str(instance.user_created.id).lower():
            raise APIException(
                self._error_while_saving
                % {"msg": "You are not the owner of this subscription"}
            )

    def check_instance_id(self, instance, validated_data):
        if validated_data["id"] and str(instance.id) != validated_data["id"]:
            raise APIException(
                self._error_while_saving % {"msg": "Invalid ID in the payload"}
            )


class SubscriptionSerializerSchema(SubscriptionSerializer):
    resourceType = fields.CharField(read_only=True)
    id = fields.UUIDField(read_only=True)
    status = fields.CharField()
    end = fields.DateTimeField()
    reason = fields.CharField()
    criteria = fields.CharField()
    channel = inline_serializer(
        name="ChannelSerializer",
        fields={
            "type": fields.CharField(),
            "endpoint": fields.CharField(),
            "header": fields.ListField(
                child=fields.CharField(), min_length=0, max_length=1
            ),
        },
    )



# Generated by Django 4.2.9 on 2024-02-07 16:47

import core.fields
import datetime
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django_cryptography.fields


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("api_fhir_r4", "0006_add_subsription_perms_imis_admin"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalsubscription",
            name="criteria",
            field=models.JSONField(blank=True, db_column="Criteria", null=True),
        ),
        migrations.AlterField(
            model_name="historicalsubscription",
            name="date_created",
            field=core.fields.DateTimeField(
                db_column="DateCreated", default=datetime.datetime.now, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubscription",
            name="date_updated",
            field=core.fields.DateTimeField(
                db_column="DateUpdated", default=datetime.datetime.now, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubscription",
            name="headers",
            field=django_cryptography.fields.encrypt(
                models.TextField(
                    blank=True, db_column="Headers", max_length=255, null=True
                )
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubscription",
            name="replacement_uuid",
            field=models.UUIDField(blank=True, db_column="ReplacementUUID", null=True),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="criteria",
            field=models.JSONField(blank=True, db_column="Criteria", null=True),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="date_created",
            field=core.fields.DateTimeField(
                db_column="DateCreated", default=datetime.datetime.now, null=True
            ),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="date_updated",
            field=core.fields.DateTimeField(
                db_column="DateUpdated", default=datetime.datetime.now, null=True
            ),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="headers",
            field=django_cryptography.fields.encrypt(
                models.TextField(
                    blank=True, db_column="Headers", max_length=255, null=True
                )
            ),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="replacement_uuid",
            field=models.UUIDField(blank=True, db_column="ReplacementUUID", null=True),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="user_created",
            field=models.ForeignKey(
                db_column="UserCreatedUUID",
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="%(class)s_user_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="user_updated",
            field=models.ForeignKey(
                db_column="UserUpdatedUUID",
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="%(class)s_user_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="subscriptionnotificationresult",
            name="error",
            field=models.TextField(blank=True, default=None, null=True),
        ),
    ]



import uuid

from policy.test_helpers import create_test_policy

from api_fhir_r4.configurations import R4IdentifierConfig
from api_fhir_r4.converters.coverageConverter import CoverageConverter
from api_fhir_r4.tests import GenericTestMixin

from api_fhir_r4.utils import TimeUtils

from core.test_helpers import create_test_officer
from insuree.test_helpers import create_test_insuree
from product.test_helpers import create_test_product


class CoverageTestMixin(GenericTestMixin):

    _TEST_POLICY_UUID = "f88687a7-1f33-466b-8c74-8b7173dc5583"
    _TEST_POLICY_ENROLL_DATE = "2021-08-20"
    _TEST_POLICY_START_DATE = "2021-08-20T00:00:00"
    _TEST_POLICY_EFFECTIVE_DATE = "2021-08-20"
    _TEST_POLICY_EXPIRED_DATE = "2022-08-19T00:00:00"
    _TEST_POLICY_STATUS = 1
    _TEST_POLICY_STAGE = "N"
    _TEST_PRODUCT_CODE = "T0001"
    _TEST_PRODUCT_NAME = "Test product"
    _TEST_INSUREE_CHFID = "chfid1"
    _TEST_PRODUCT_UUID = "8ed8d2d9-2644-4d29-ba37-ab772386cfca"

    _TEST_POLICY = None
    _TEST_POLICY_VALUE = 10000.0
    test_insuree = None
    sub_str = {}

    def setUp(self):
        super(GenericTestMixin, self).setUp()
        self.create_dependencies()
        self.sub_str[self._TEST_INSUREE_CHFID] = self.test_insuree.chf_id
        self.sub_str[self._TEST_PRODUCT_UUID] = self._TEST_PRODUCT.uuid

    @classmethod
    def create_dependencies(cls):
        # create mocked insuree
        cls.test_insuree = create_test_insuree(
            with_family=True, custom_props={"chf_id": cls._TEST_INSUREE_CHFID}
        )
        imis_officer = create_test_officer()
        cls._TEST_PRODUCT = create_test_product(
            cls._TEST_PRODUCT_CODE, valid=True, custom_props={}
        )

        cls._TEST_POLICY = create_test_policy(
            product=cls._TEST_PRODUCT,
            insuree=cls.test_insuree,
            custom_props={
                "uuid": cls._TEST_POLICY_UUID,
                "officer_id": imis_officer.id,
                "family": cls.test_insuree.family,
                "enroll_date": TimeUtils.str_to_date(cls._TEST_POLICY_ENROLL_DATE),
                "effective_date": TimeUtils.str_to_date(
                    cls._TEST_POLICY_EFFECTIVE_DATE
                ),
                "expiry_date": TimeUtils.str_to_date(cls._TEST_POLICY_EXPIRED_DATE),
                "stage": cls._TEST_POLICY_STAGE,
                "status": cls._TEST_POLICY_STATUS,
                "value": cls._TEST_POLICY_VALUE,
            },
        )

    def create_test_imis_instance(self):
        return self._TEST_POLICY

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(
            self._TEST_POLICY_ENROLL_DATE, imis_obj.enroll_date.isoformat()
        )
        self.assertEqual(self._TEST_POLICY_START_DATE, imis_obj.start_date.isoformat())
        self.assertEqual(
            self._TEST_POLICY_EFFECTIVE_DATE, imis_obj.effective_date.isoformat()
        )
        self.assertEqual(
            self._TEST_POLICY_EXPIRED_DATE, imis_obj.expiry_date.isoformat()
        )
        self.assertEqual(self._TEST_PRODUCT_CODE, imis_obj.product.code)
        self.assertEqual(self._TEST_PRODUCT_UUID, str(uuid.UUID(imis_obj.product.uuid)))

    def verify_fhir_instance(self, fhir_obj):
        for identifier in fhir_obj.identifier:
            code = CoverageConverter.get_first_coding_from_codeable_concept(
                identifier.type
            ).code
            if code == R4IdentifierConfig.get_fhir_uuid_type_code():
                self.assertEqual(self._TEST_POLICY_UUID, identifier.value)
        self.assertIn(
            fhir_obj.policyHolder.reference, f"Patient/{self.test_insuree.chf_id}"
        )
        self.assertIn(
            fhir_obj.beneficiary.reference, f"Patient/{self.test_insuree.chf_id}"
        )
        self.assertIn(
            fhir_obj.payor[0].reference, f"Patient/{self.test_insuree.chf_id}"
        )
        self.assertEqual(1, len(fhir_obj.class_fhir))
        self.assertEqual("plan", fhir_obj.class_fhir[0].type.coding[0].code)
        self.assertEqual(self._TEST_PRODUCT_CODE, fhir_obj.class_fhir[0].value)
        self.assertEqual(
            f"{self._TEST_PRODUCT_NAME} {self._TEST_PRODUCT_CODE}",
            fhir_obj.class_fhir[0].name,
        )
        period = fhir_obj.period
        self.assertEqual(
            self._TEST_POLICY_START_DATE[:10], period.start.isoformat()[:10]
        )
        self.assertEqual(
            self._TEST_POLICY_EXPIRED_DATE[:10], period.end.isoformat()[:10]
        )




# flake8: noqa

from api_fhir_r4.converters.baseFHIRConverter import BaseFHIRConverter
from api_fhir_r4.converters.personConverterMixin import PersonConverterMixin
from api_fhir_r4.converters.referenceConverterMixin import ReferenceConverterMixin
from api_fhir_r4.converters.medicationConverter import MedicationConverter
from api_fhir_r4.converters.activityDefinitionConverter import ActivityDefinitionConverter
from api_fhir_r4.converters.contractConverter import ContractConverter
from api_fhir_r4.converters.patientConverter import PatientConverter
from api_fhir_r4.converters.groupConverter import GroupConverter
from api_fhir_r4.converters.policyHolderOrganisationConverter import PolicyHolderOrganisationConverter
from api_fhir_r4.converters.locationConverter import LocationConverter
from api_fhir_r4.converters.locationSiteConverter import LocationSiteConverter
from api_fhir_r4.converters.operationOutcomeConverter import OperationOutcomeConverter
from api_fhir_r4.converters.claimAdminPractitionerConverter import ClaimAdminPractitionerConverter
from api_fhir_r4.converters.claimAdminPractitionerRoleConverter import ClaimAdminPractitionerRoleConverter
from api_fhir_r4.converters.coverageEligibilityRequestConverter import CoverageEligibilityRequestConverter
from api_fhir_r4.converters.healthFacilityOrganisationConverter import HealthFacilityOrganisationConverter
from api_fhir_r4.converters.claimConverter import ClaimConverter
from api_fhir_r4.converters.claimResponseConverter import ClaimResponseConverter
from api_fhir_r4.converters.codeSystemConverter import CodeSystemConverter
from api_fhir_r4.converters.coverageConverter import CoverageConverter
from api_fhir_r4.converters.enrolmentOfficerPractitionerConverter import EnrolmentOfficerPractitionerConverter
from api_fhir_r4.converters.enrolmentOfficerPractitionerRoleConverter import EnrolmentOfficerPractitionerRoleConverter
from api_fhir_r4.converters.groupConverter import GroupConverter
from api_fhir_r4.converters.insuranceOrganisationConverter import InsuranceOrganisationConverter
from api_fhir_r4.converters.insurancePlanConverter import InsurancePlanConverter
from api_fhir_r4.converters.invoiceConverter import GenericInvoiceConverter
from api_fhir_r4.converters.invoiceConverter import InvoiceConverter
from api_fhir_r4.converters.invoiceConverter import BillInvoiceConverter
from api_fhir_r4.converters.communicationRequestConverter import CommunicationRequestConverter

from api_fhir_r4.converters.policyCoverageEligibilityRequestConverter import PolicyCoverageEligibilityRequestConverter
from api_fhir_r4.converters.communicationConverter import CommunicationConverter
from api_fhir_r4.converters.valueSetConverter import ValueSetConverter



from api_fhir_r4.converters import ContractConverter
from api_fhir_r4.tests import ContractTestMixin
from fhir.resources.R4B.contract import Contract

from api_fhir_r4.tests.mixin import (
    ConvertToImisTestMixin,
    ConvertToFhirTestMixin,
    ConvertJsonToFhirTestMixin,
)


class ContractConverterTestCase(
    ContractTestMixin,
    ConvertToImisTestMixin,
    ConvertToFhirTestMixin,
    ConvertJsonToFhirTestMixin,
):
    converter = ContractConverter
    fhir_resource = Contract
    json_repr = "test/test_contract.json"



from abc import ABC, abstractmethod
from datetime import timedelta

from django.db.models import QuerySet
from typing import Dict, Callable, Any

from core.datetimes.ad_datetime import datetime


class QuerysetFilterABC(ABC):
    def __init__(self, field, value):
        self.field = field
        self.value = value

    @abstractmethod
    def apply_filter(self, queryset: QuerySet) -> QuerySet:
        """
        apply_filter should apply contained filter to given queryset and return the result.
        @param queryset: queryset the filter should be applied to
        @return: filtered queryset
        """
        pass


class QuerysetEqualFilter(QuerysetFilterABC):
    def apply_filter(self, queryset):
        return queryset.filter(**{self.field: self.value})


class QuerysetNotEqualFilter(QuerysetFilterABC):
    def apply_filter(self, queryset):
        return queryset.exclude(**{self.field: self.value})


class QuerysetGreaterThanFilter(QuerysetFilterABC):
    def apply_filter(self, queryset):
        return queryset.filter(**{f"{self.field}__gt": self.value})


class QuerysetLesserThanFilter(QuerysetFilterABC):
    def apply_filter(self, queryset):
        return queryset.filter(**{f"{self.field}__lt": self.value})


class QuerysetGreaterThanEqualFilter(QuerysetFilterABC):
    def apply_filter(self, queryset):
        return queryset.filter(**{f"{self.field}__gte": self.value})


class QuerysetLesserThanEqualFilter(QuerysetFilterABC):
    def apply_filter(self, queryset):
        return queryset.filter(**{f"{self.field}__lte": self.value})


class QuerysetApproximateDateFilter(QuerysetFilterABC):
    def apply_filter(self, queryset):
        range_size = (datetime.now() - self.value).days * 0.1
        value_start = self.value - timedelta(days=range_size)
        value_end = self.value + timedelta(days=range_size)
        return queryset.filter(**{f"{self.field}__range": (value_start, value_end)})


class QuerysetParameterABC(ABC):
    def __init__(self, output_parameter):
        self.output_parameter = output_parameter
        self.accepted_prefixes = self._get_prefix_filter_mapping().keys()

    @abstractmethod
    def _get_prefix_filter_mapping(
        self,
    ) -> Dict[str, Callable[[str, Any], QuerysetFilterABC]]:
        """
        _get_prefix_filter_mapping should return a dict that maps respective prefixes from FHIR specification to lambdas
        capable of creating specific filters taking affected field and parsed parameter value as arguments.
        @return: {prefix: lambda creating filter} map
        """

    def build_filter(self, request_parameter_value):
        modifier, value = self._get_prefix_and_value(request_parameter_value)
        return self._get_prefix_filter_mapping()[modifier if modifier else "eq"](
            self.output_parameter, value
        )

    def _get_prefix_and_value(self, parameter):
        modifier = next(
            (
                modifier
                for modifier in self.accepted_prefixes
                if parameter.startswith(modifier)
            ),
            "",
        )
        output_value = self._parse_value(parameter[len(modifier):])
        return modifier, output_value

    def _parse_value(self, value):
        """
        Allow for custom parameter value parsing logic. _parse_value in case of parsing error should raise Value error
        with message containing {request_parameter} placeholder to insert parameter name
        @param value: value to be parsed
        @return: parsed value
        """
        return value


class QuerysetLastUpdatedParameter(QuerysetParameterABC):
    def _get_prefix_filter_mapping(self):
        return {
            "eq": lambda field, value: QuerysetEqualFilter(field, value),
            "ne": lambda field, value: QuerysetNotEqualFilter(field, value),
            "gt": lambda field, value: QuerysetGreaterThanFilter(field, value),
            "lt": lambda field, value: QuerysetLesserThanFilter(field, value),
            "ge": lambda field, value: QuerysetGreaterThanEqualFilter(field, value),
            "le": lambda field, value: QuerysetLesserThanEqualFilter(field, value),
            "sa": lambda field, value: QuerysetGreaterThanEqualFilter(field, value),
            "eb": lambda field, value: QuerysetLesserThanEqualFilter(field, value),
            "ap": lambda field, value: QuerysetApproximateDateFilter(field, value),
        }

    def _parse_value(self, value):
        try:
            return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S")
        except Exception:
            raise ValueError("{request_parameter} value is not a valid datetime")


class RequestParameterFilterABC(ABC):
    def __init__(self, request):
        self.request = request

    @abstractmethod
    def _get_parameter_mapping(self) -> Dict[str, Callable[[], QuerysetParameterABC]]:
        """
        _get_parameter_mapping should return a dict mapping request parameters to lambdas capable of creating
        filters (allowing lazy loading)
        @return: {request parameter: lambda creating queryset parameter} map
        """

    def filter_queryset(self, queryset):
        parameter_mapping = self._get_parameter_mapping()
        accepted_parameters = parameter_mapping.keys()
        request_parameters = {
            parameter: self.request.GET[parameter]
            for parameter in accepted_parameters
            if parameter in self.request.GET
        }

        output_queryset = queryset
        for request_parameter in request_parameters:
            output_parameter = parameter_mapping[request_parameter]()
            try:
                output_queryset = output_parameter.build_filter(
                    self.request.GET[request_parameter]
                ).apply_filter(output_queryset)
            except ValueError as parsingError:
                raise ValueError(
                    str(parsingError).format(**{"request_parameter": request_parameter})
                )

        return output_queryset


class ValidityFromRequestParameterFilter(RequestParameterFilterABC):
    def _get_parameter_mapping(self):
        return {
            "_lastUpdated": lambda: QuerysetLastUpdatedParameter("validity_from"),
        }


class DateUpdatedRequestParameterFilter(RequestParameterFilterABC):
    def _get_parameter_mapping(self):
        return {
            "_lastUpdated": lambda: QuerysetLastUpdatedParameter("date_updated"),
        }



from api_fhir_r4.converters.contractConverter import ContractConverter
from api_fhir_r4.serializers.baseSerializer import BaseFHIRSerializer
from contribution.services import update_or_create_premium
from policy.services import PolicyService
from policy.models import Policy
import copy
from api_fhir_r4.exceptions import FHIRException


class ContractSerializer(BaseFHIRSerializer):
    fhirConverter = ContractConverter

    def create(self, validated_data):
        request = self.context.get("request")
        user = request.user

        family = validated_data.get("family_id")
        insurees = validated_data.pop("insurees")
        premiums = validated_data.pop("contributions")

        if (
            Policy.objects.filter(family_id=family)
            .filter(
                start_date__range=[
                    validated_data.get("effective_date"),
                    validated_data.get("expiry_date"),
                ]
            )
            .count()
            > 0
        ):
            raise FHIRException("Contract exists for this patient")

        copied_data = copy.deepcopy(validated_data)
        if "_state" in copied_data:
            del copied_data["_state"]
        # TODO should we implement a way to create a resource with a given uuid
        del copied_data["uuid"]

        # Add insurees to the data for PolicyService
        if insurees:
            copied_data["members_uuid"] = insurees

        new_policy = PolicyService(user).update_or_create(copied_data, user)
        # create contributions related to newly created policy
        if premiums:
            for premium in premiums:
                premium.policy = new_policy
                premium = update_or_create_premium(premium, user)
                new_policy = premium.policy

        return new_policy

    def update(self, instance, validated_data):
        instance.save()
        return instance



from api_fhir_r4.configurations import CommunicationRequestConfiguration


class R4CommunicationRequestConfig(CommunicationRequestConfiguration):

    @classmethod
    def build_configuration(cls, cfg):
        cls.get_config().R4_fhir_communication_config = cfg[
            "R4_fhir_communication_request_config"
        ]

    @classmethod
    def get_fhir_care_rendered_code(cls):
        return cls.get_config_attribute("R4_fhir_communication_config").get(
            "fhir_care_rendered_code", "CareRendered"
        )

    @classmethod
    def get_fhir_payment_asked_code(cls):
        return cls.get_config_attribute("R4_fhir_communication_config").get(
            "fhir_payment_asked_code", "PaymentAsked"
        )

    @classmethod
    def get_fhir_drug_prescribed_code(cls):
        return cls.get_config_attribute("R4_fhir_communication_config").get(
            "fhir_drug_prescribed_code", "DrugPrescribed"
        )

    @classmethod
    def get_fhir_drug_received_code(cls):
        return cls.get_config_attribute("R4_fhir_communication_config").get(
            "fhir_drug_received_code", "DrugReceived"
        )

    @classmethod
    def get_fhir_asessment_code(cls):
        return cls.get_config_attribute("R4_fhir_communication_config").get(
            "fhir_asessment_code", "Asessment"
        )



from api_fhir_r4.configurations import MaritalConfiguration


class R4MaritalConfig(MaritalConfiguration):

    @classmethod
    def build_configuration(cls, cfg):
        cls.get_config().R4_fhir_marital_status_config = cfg[
            "R4_fhir_marital_status_config"
        ]

    @classmethod
    def get_fhir_marital_status_system(cls):
        return cls.get_config_attribute("R4_fhir_marital_status_config").get(
            "system", "http://hl7.org/fhir/valueset-marital-status.html"
        )

    @classmethod
    def get_fhir_married_code(cls):
        return cls.get_config_attribute("R4_fhir_marital_status_config").get(
            "fhir_code_for_married", "M"
        )

    @classmethod
    def get_fhir_never_married_code(cls):
        return cls.get_config_attribute("R4_fhir_marital_status_config").get(
            "fhir_code_for_never_married", "S"
        )

    @classmethod
    def get_fhir_divorced_code(cls):
        return cls.get_config_attribute("R4_fhir_marital_status_config").get(
            "fhir_code_for_divorced", "D"
        )

    @classmethod
    def get_fhir_widowed_code(cls):
        return cls.get_config_attribute("R4_fhir_marital_status_config").get(
            "fhir_code_for_widowed", "W"
        )

    @classmethod
    def get_fhir_unknown_marital_status_code(cls):
        return cls.get_config_attribute("R4_fhir_marital_status_config").get(
            "fhir_code_for_unknown", "U"
        )



from rest_framework.viewsets import GenericViewSet

from api_fhir_r4.mixins import MultiIdentifierRetrieverMixin, ListModelMixin
from api_fhir_r4.model_retrievers import (
    UUIDIdentifierModelRetriever,
    CodeIdentifierModelRetriever,
)
from api_fhir_r4.permissions import FHIRApiActivityDefinitionPermissions
from api_fhir_r4.serializers import ActivityDefinitionSerializer
from api_fhir_r4.views.fhir.base import BaseFHIRView
from api_fhir_r4.views.filters import ValidityFromRequestParameterFilter
from medical.models import Service


class ActivityDefinitionViewSet(
    BaseFHIRView, MultiIdentifierRetrieverMixin, ListModelMixin, GenericViewSet
):
    retrievers = [UUIDIdentifierModelRetriever, CodeIdentifierModelRetriever]
    serializer_class = ActivityDefinitionSerializer
    permission_classes = (FHIRApiActivityDefinitionPermissions,)

    def get_queryset(self):
        queryset = Service.get_queryset(None, self.request.user)
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(
            queryset
        )



import logging

from django.db import migrations
from django.db.models import Q

logger = logging.getLogger(__name__)


def update_subscription_criteria(apps, schema_editor):
    resource_type_map = {
        "Invoice": "Invoice",
        "Bill": "Invoice",
        "Patient": "Patient",
        "Organisation": "Organisation",
    }

    subscription_model = apps.get_model("api_fhir_r4", "Subscription")
    entries_to_update = subscription_model.objects.filter(
        Q(criteria__has_key="resource_type") & ~Q(criteria__has_key="resource")
    )
    for subscription in entries_to_update:
        resource = subscription.criteria.pop("resource_type")
        if resource in resource_type_map:
            subscription.criteria["resource"] = resource_type_map[resource]
            subscription.save()
        else:
            logger.warning(f"Subscription of unexpected type - {resource}")


class Migration(migrations.Migration):

    dependencies = [
        ("api_fhir_r4", "0003_auto_20220313_1634"),
    ]

    operations = [
        migrations.RunPython(update_subscription_criteria, migrations.RunPython.noop)
    ]



from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig
from api_fhir_r4.converters import (
    BaseFHIRConverter,
    EnrolmentOfficerPractitionerConverter,
    ReferenceConverterMixin,
)
from api_fhir_r4.converters.healthFacilityOrganisationConverter import (
    LocationConverter,
    PersonConverterMixin,
)
from api_fhir_r4.utils import DbManagerUtils
from core.models import Officer
from django.utils.translation import gettext as _
from fhir.resources.R4B.extension import Extension
from fhir.resources.R4B.practitionerrole import PractitionerRole


class EnrolmentOfficerPractitionerRoleConverter(
    BaseFHIRConverter, PersonConverterMixin, ReferenceConverterMixin
):

    @classmethod
    def to_fhir_obj(
        cls, imis_officer, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE
    ):
        fhir_practitioner_role = PractitionerRole.construct()
        cls.build_fhir_pk(fhir_practitioner_role, imis_officer, reference_type)
        cls.build_fhir_identifiers(fhir_practitioner_role, imis_officer)
        cls.build_fhir_extension(fhir_practitioner_role, imis_officer, reference_type)
        cls.build_fhir_practitioner_reference(
            fhir_practitioner_role, imis_officer, reference_type
        )
        cls.build_fhir_location_references(
            fhir_practitioner_role, imis_officer, reference_type
        )
        cls.build_fhir_code(fhir_practitioner_role)
        cls.build_fhir_telecom(fhir_practitioner_role, imis_officer)
        return fhir_practitioner_role

    @classmethod
    def to_imis_obj(cls, fhir_practitioner_role, audit_user_id):
        errors = []
        fhir_practitioner_role = PractitionerRole(**fhir_practitioner_role)
        practitioner = fhir_practitioner_role.practitioner
        imis_officer = (
            EnrolmentOfficerPractitionerConverter.get_imis_obj_by_fhir_reference(
                practitioner, errors
            )
        )
        location_references = fhir_practitioner_role.location
        if len(location_references) != 1:
            errors.append("Location does not have 1 element")
        else:
            location = cls.get_location_by_reference(location_references[0], errors)
            substitution_officer = None
        if (
            fhir_practitioner_role.extension
            and len(fhir_practitioner_role.extension) > 0
        ):
            substitution_officer_reference = fhir_practitioner_role.extension[
                0
            ].valueReference
            substitution_officer = (
                EnrolmentOfficerPractitionerConverter.get_imis_obj_by_fhir_reference(
                    substitution_officer_reference,
                )
            )
        if not cls.valid_condition(
            imis_officer is None, "Practitioner doesn't exists", errors
        ):
            imis_officer.location = location
            if substitution_officer:
                imis_officer.substitution_officer = substitution_officer
        cls.check_errors(errors)
        return imis_officer

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_generic_type_code()

    @classmethod
    def get_reference_obj_uuid(cls, imis_officer):
        return imis_officer.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_officer):
        return imis_officer.id

    @classmethod
    def get_reference_obj_code(cls, imis_officer):
        return imis_officer.code

    @classmethod
    def get_fhir_resource_type(cls):
        return PractitionerRole

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        return DbManagerUtils.get_object_or_none(
            Officer, **cls.get_database_query_id_parameteres_from_reference(reference)
        )

        imis_officer_code = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Officer, code=imis_officer_code)
        return DbManagerUtils.get_object_or_none(
            Officer, **cls.get_database_query_id_parameteres_from_reference(reference)
        )

    @classmethod
    def build_fhir_extension(cls, fhir_practitioner_role, imis_officer, reference_type):
        if imis_officer.substitution_officer:
            extension = Extension.construct()
            extension.url = f"""{
                GeneralConfiguration.get_system_base_url()
            }StructureDefinition/practitioner-role-substitution-reference"""

            reference = (
                EnrolmentOfficerPractitionerConverter.build_fhir_resource_reference(
                    imis_officer.substitution_officer, reference_type=reference_type
                )
            )
            extension.valueReference = reference
            fhir_practitioner_role.extension = [extension]

    @classmethod
    def build_fhir_identifiers(cls, fhir_practitioner_role, imis_officer):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_officer)
        fhir_practitioner_role.identifier = identifiers

    @classmethod
    def build_fhir_code_identifier(cls, identifiers, imis_officer):
        if imis_officer.code:
            identifier = cls.build_fhir_identifier(
                imis_officer.code,
                R4IdentifierConfig.get_fhir_identifier_type_system(),
                R4IdentifierConfig.get_fhir_generic_type_code(),
            )
            identifiers.append(identifier)

    @classmethod
    def build_fhir_practitioner_reference(
        cls, fhir_practitioner_role, imis_officer, reference_type
    ):
        fhir_practitioner_role.practitioner = (
            EnrolmentOfficerPractitionerConverter.build_fhir_resource_reference(
                imis_officer, reference_type=reference_type
            )
        )

    @classmethod
    def build_fhir_location_references(
        cls, fhir_practitioner_role, imis_officer, reference_type
    ):
        if imis_officer.location:
            reference = LocationConverter.build_fhir_resource_reference(
                imis_officer.location, reference_type=reference_type
            )
            fhir_practitioner_role.location = [reference]

    @classmethod
    def get_location_by_reference(cls, location_references, errors):
        return LocationConverter.get_imis_obj_by_fhir_reference(location_references)

    @classmethod
    def get_first_location(cls, location_references):
        return location_references[0]

    @classmethod
    def build_fhir_code(cls, fhir_practitioner_role):
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/practitioner-qualification-type"
        fhir_practitioner_role.code = [
            cls.build_codeable_concept(
                system=system, code="EO", display=_("Enrolment Officer")
            )
        ]

    @classmethod
    def build_fhir_telecom(cls, fhir_practitioner_role, imis_officer):
        fhir_practitioner_role.telecom = cls.build_fhir_telecom_for_person(
            phone=imis_officer.phone, email=imis_officer.email
        )



from api_fhir_r4.converters import BaseFHIRConverter
from fhir.resources.R4B.valueset import ValueSet, ValueSetCompose, ValueSetComposeInclude, ValueSetExpansion, ValueSetExpansionContains
from api_fhir_r4.utils import FhirUtils


class ValueSetConverter(BaseFHIRConverter):

    @classmethod
    def to_imis_obj(cls, data, audit_user_id):
        raise NotImplementedError("`toImisObj()` not implemented.")  # pragma: no cover

    @classmethod
    def get_fhir_code_identifier_type(cls):
        raise NotImplementedError(
            "`get_fhir_code_identifier_type()` not implemented."
        )  # pragma: no cover

    @classmethod
    def to_fhir_obj(cls, obj, reference_type):
        fhir_value_set = {}
        cls.build_fhir_value_set_status(fhir_value_set)
        cls.build_fhir_value_set_compose(fhir_value_set, obj)
        cls.build_fhir_value_set_expansion(fhir_value_set, obj)
        fhir_value_set = ValueSet(**fhir_value_set)
        cls.build_fhir_id(fhir_value_set, obj)
        cls.build_fhir_url(fhir_value_set, obj)
        cls.build_fhir_value_set_name(fhir_value_set, obj)
        cls.build_fhir_value_set_title(fhir_value_set, obj)
        cls.build_fhir_value_set_date(fhir_value_set)
        cls.build_fhir_value_set_description(fhir_value_set, obj)
        return fhir_value_set

    @classmethod
    def build_fhir_id(cls, fhir_value_set, obj):
        fhir_value_set.id = obj["id"]

    @classmethod
    def build_fhir_url(cls, fhir_value_set, obj):
        fhir_value_set.url = obj["url"]

    @classmethod
    def build_fhir_value_set_name(cls, fhir_value_set, obj):
        fhir_value_set.name = obj["name"]

    @classmethod
    def build_fhir_value_set_title(cls, fhir_value_set, obj):
        fhir_value_set.title = obj["title"]

    @classmethod
    def build_fhir_value_set_date(cls, fhir_value_set):
        from core.utils import TimeUtils

        fhir_value_set.date = TimeUtils.now()

    @classmethod
    def build_fhir_value_set_description(cls, fhir_value_set, obj):
        fhir_value_set.description = obj["description"]

    @classmethod
    def build_fhir_value_set_status(cls, fhir_value_set):
        fhir_value_set["status"] = "active"

    @classmethod
    def build_fhir_value_set_compose(cls, fhir_value_set, obj):
        compose = ValueSetCompose.construct()
        compose.include = []

        include = ValueSetComposeInclude.construct()
        include.system = obj["codesystem_url"]
        compose.include.append(include)

        fhir_value_set.compose = compose

    @classmethod
    def build_fhir_value_set_expansion(cls, fhir_value_set, obj):
        expansion = ValueSetExpansion.construct()
        expansion.contains = []

        for item in obj["data"]:
            contains = ValueSetExpansionContains.construct()
            contains.code = FhirUtils.get_attr(item, obj["code_field"])
            contains.display = FhirUtils.get_attr(item, obj["display_field"])

            # Add properties for price and other parameters
            if obj.get("properties"):
                contains.property = []
                for prop_name in obj["properties"]:
                    if prop_name in item:
                        prop = {
                            "code": prop_name,
                            "valueString": str(item[prop_name])
                        }
                        contains.property.append(prop)

            expansion.contains.append(contains)

        expansion.total = len(expansion.contains)
        fhir_value_set.expansion = expansion




from rest_framework import status
from rest_framework.test import APITestCase

from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.tests import GenericFhirAPITestMixin
from api_fhir_r4.tests.mixin.logInMixin import LogInMixin
from api_fhir_r4.utils import TimeUtils
from api_fhir_r4.tests.utils import (
    load_and_replace_json,
    get_connection_payload,
    get_or_create_user_api,
)
from core.test_helpers import create_admin_role
from claim.models import Claim
from location.models import UserDistrict
from claim.test_helpers import create_test_claim_context, full_delete_claim
from datetime import datetime


class ClaimAPITests(GenericFhirAPITestMixin, APITestCase, LogInMixin):
    base_url = GeneralConfiguration.get_base_url() + "Claim/"
    _test_json_path = "/test/test_claim.json"

    # diagnosis data
    _TEST_MAIN_ICD_CODE = "T_CD"
    _TEST_MAIN_ICD_NAME = "Test diagnosis"

    _TEST_CLAIM_ADMIN_UUID = "044c33d1-dbf3-4d6a-9924-3797b461e535"
    _TEST_CLAIM_ADMIN_CODE = "VITEST99"
    _TEST_INSUREE_CODE = "999000001"
    _TEST_INSUREE_UUID = "76aca309-f8cf-4890-8f2e-b416d78de00b"
    # claim item data
    _TEST_ITEM_CODE = "0004"
    _TEST_ITEM_UUID = "e2bc1546-390b-4d41-8571-632ecf7a0936"
    _TEST_ITEM_QUANTITY_PROVIDED = 10.0
    _TEST_ITEM_PRICE_ASKED = 10.0
    _TEST_ITEM_EXPLANATION = "item_explanation"
    _TEST_ITEM_TYPE = "D"

    # claim service data
    _TEST_SERVICE_CODE = "M7"
    _TEST_SERVICE_UUID = "a17602f4-e9ff-4f42-a6a4-ccefcb23b4d6"
    _TEST_SERVICE_QUANTITY_PROVIDED = 1
    _TEST_SERVICE_PRICE_ASKED = 21000.0
    _TEST_SERVICE_EXPLANATION = "service_explanation"
    _TEST_SERVICE_TYPE = "D"

    # hf test data
    _TEST_HF_ID = 10000
    _TEST_HF_UUID = "6d0eea8c-62eb-11ea-94d6-c36229a16c2f"
    _TEST_HF_CODE = "12345678"
    _TEST_HF_NAME = "TEST_NAME"
    _TEST_HF_LEVEL = "H"
    _TEST_HF_LEGAL_FORM = "G"
    _TEST_ADDRESS = "TEST_ADDRESS"
    _TEST_PHONE = "133-996-476"
    _TEST_FAX = "1-408-999 8888"
    _TEST_EMAIL = "TEST@TEST.com"
    _TEST_DATE = "2021-02-03"

    _ADMIN_AUDIT_USER_ID = -1
    _TEST_USER = None
    _TEST_USER_NAME = "Admin"
    _TEST_USER_PASSWORD = "admin123"
    _TEST_DATA_USER = {
        "username": _TEST_USER_NAME,
        "last_name": _TEST_USER_NAME,
        "password": _TEST_USER_PASSWORD,
        "other_names": _TEST_USER_NAME,
        "user_types": "INTERACTIVE",
        "language": "en",
        "roles": [create_admin_role().id],
    }

    _test_json_path_credentials = "/test/test_login.json"
    _test_request_data_credentials = None
    _test_json_path_with_code_references = "/test/test_claim_with_code_references.json"
    sub_str = {}
    test_insuree = None
    test_claim_admin = None

    def setUp(self):
        super(ClaimAPITests, self).setUp()
        self._TEST_USER = get_or_create_user_api(self._TEST_DATA_USER)
        self.create_dependencies()
        self.sub_str[self._TEST_INSUREE_UUID] = self.test_insuree.uuid
        self.sub_str[self._TEST_INSUREE_CODE] = self.test_insuree.chf_id
        self.sub_str[self._TEST_HF_UUID] = self.test_hf.uuid
        self.sub_str[self._TEST_CLAIM_ADMIN_UUID] = self.test_claim_admin.uuid
        self.sub_str[self._TEST_CLAIM_ADMIN_CODE] = self.test_claim_admin.code

        self.sub_str[self._TEST_ITEM_UUID] = self._TEST_ITEM.uuid
        self.sub_str[self._TEST_SERVICE_UUID] = self._TEST_SERVICE.uuid
        self.sub_str[self._TEST_ITEM_CODE] = self._TEST_ITEM.code
        self.sub_str[self._TEST_SERVICE_CODE] = self._TEST_SERVICE.code
        self.sub_str[self._TEST_DATE] = datetime.now().strftime("%Y-%m-%d")

        self._TEST_HF_ID = self.test_hf.id
        self._TEST_HF_UUID = self.test_hf.uuid

    def create_dependencies(self):

        self.test_claim, self.test_insuree, self.policy, self.test_hf = (
            create_test_claim_context(
                claim={
                    "icd": {
                        "code": self._TEST_MAIN_ICD_CODE,
                        "name": self._TEST_MAIN_ICD_NAME,
                    }
                },
                insuree={"chf_id": self._TEST_INSUREE_CODE},
                claim_admin={
                    "code": "T-CA-API",
                    "last_name": self._TEST_DATA_USER["last_name"],
                    "other_names": self._TEST_DATA_USER["other_names"],
                },
                product={},
                hf={
                    "name": self._TEST_HF_NAME,
                    "code": self._TEST_HF_CODE,
                    "level": self._TEST_HF_LEVEL,
                    "legal_form_id": self._TEST_HF_LEGAL_FORM,
                    "address": self._TEST_ADDRESS,
                    "phone": self._TEST_PHONE,
                    "fax": self._TEST_FAX,
                    "email": self._TEST_EMAIL,
                },
                items=[{"type": self._TEST_ITEM_TYPE}],
                services=[{"type": self._TEST_SERVICE_TYPE}],
            )
        )
        #
        self.test_claim_admin = self.test_claim.admin
        self._TEST_ITEM = self.test_claim.items.first().item
        self._TEST_SERVICE = self.test_claim.services.first().service

        full_delete_claim(self.test_claim.id)

        ud = UserDistrict()
        ud.location = self.test_insuree.family.location.parent.parent
        ud.audit_user_id = self._ADMIN_AUDIT_USER_ID
        ud.user = self._TEST_USER.i_user
        ud.validity_from = TimeUtils.now()
        ud.save()

    def _post_claim(self, data, headers):
        return self.client.post(self.base_url, data=data, format="json", **headers)

    def test_post_should_create_correctly(self):
        response = self.client.post(
            GeneralConfiguration.get_base_url() + "login/",
            data=get_connection_payload(self._TEST_DATA_USER),
            format="json",
        )
        response_json = response.json()
        token = response_json["token"]
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        headers = {
            "Content-Type": "application/json",
            "HTTP_AUTHORIZATION": f"Bearer {token}",
        }

        dataset = [
            load_and_replace_json(self._test_json_path, self.sub_str),
            load_and_replace_json(
                self._test_json_path_with_code_references, self.sub_str
            ),
        ]
        for data in dataset:
            response = self.client.post(
                self.base_url, data=data, format="json", **headers
            )
            self.assertEqual(
                response.status_code, status.HTTP_201_CREATED, response.json()
            )
            self.assertIsNotNone(response.content)
            response_json = response.json()
            self.assertEqual(response_json["resourceType"], "ClaimResponse")
            self.assertEqual(response_json["outcome"], "complete")
            # for both tests item and service should be 'rejected' and 'not in price list'
            for item in response_json["item"]:
                for adjudication in item["adjudication"]:
                    self.assertEqual(
                        adjudication["category"]["coding"][0]["code"],
                        f"{Claim.STATUS_REJECTED}",
                    )
                    # 2 not in price list
                    self.assertEqual(adjudication["reason"]["coding"][0]["code"], "2")

    def test_get_should_return_200_claim_response(self):
        # test if get ClaimResponse return 200
        response = self.client.post(
            GeneralConfiguration.get_base_url() + "login/",
            data=get_connection_payload(self._TEST_DATA_USER),
            format="json",
        )
        response_json = response.json()
        token = response_json["token"]
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        headers = {
            "Content-Type": "application/json",
            "HTTP_AUTHORIZATION": f"Bearer {token}",
        }
        response = self.client.get(
            GeneralConfiguration.get_base_url() + "ClaimResponse/",
            data=None,
            format="json",
            **headers,
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)



from abc import ABC
from insuree.test_helpers import (
    create_test_gender,
    create_test_profession,
    create_test_education,
    create_test_relation,
    create_test_confirmation_type
)
from location.test_helpers import (
    create_basic_test_locations,
    create_test_basic_health_facility_legal_form,
    create_test_basic_health_facility_sub_level
)

from django.test import TestCase


class GenericTestMixin(TestCase, ABC):  # pragma: no cover

    @classmethod
    def setUpTestData(cls):
        super(GenericTestMixin, cls).setUpTestData()
        create_basic_test_locations()
        create_test_basic_health_facility_legal_form()
        create_test_basic_health_facility_sub_level()
        create_test_gender()
        create_test_profession()
        create_test_education()
        create_test_relation()
        create_test_confirmation_type()

    def create_test_imis_instance(self):
        raise NotImplementedError("`test_imis_instance()` must be implemented.")

    def verify_imis_instance(self, imis_obj):
        raise NotImplementedError("`verify_imis_instance()` must be implemented.")

    def create_test_fhir_instance(self):
        raise NotImplementedError("`test_fhir_instance()` must be implemented.")

    def verify_fhir_instance(self, fhir_obj):
        raise NotImplementedError("`verify_fhir_instance()` must be implemented.")



from drf_spectacular.utils import extend_schema, extend_schema_view
from rest_framework.exceptions import PermissionDenied, APIException
from rest_framework.viewsets import ModelViewSet

from api_fhir_r4.models import Subscription
from api_fhir_r4.openapi_schema_extensions import get_inline_error_serializer
from api_fhir_r4.permissions import FHIRApiSubscriptionPermissions
from api_fhir_r4.subscriptions import SubscriptionSerializer
from api_fhir_r4.services import SubscriptionService
from api_fhir_r4.subscriptions.subscriptionSerializer import (
    SubscriptionSerializerSchema,
)
from api_fhir_r4.views.fhir.base import BaseFHIRView
from api_fhir_r4.views.filters import DateUpdatedRequestParameterFilter
from api_fhir_r4.mixins import ListModelMixin, RetrieveModelMixin


@extend_schema_view(
    list=extend_schema(
        responses={(200, "application/json"): SubscriptionSerializerSchema()}
    ),
    retrieve=extend_schema(
        responses={
            (200, "application/json"): SubscriptionSerializerSchema(),
            (404, "application/json"): get_inline_error_serializer(),
        }
    ),
    create=extend_schema(
        request=SubscriptionSerializerSchema(),
        responses={(201, "application/json"): SubscriptionSerializerSchema()},
    ),
    update=extend_schema(
        request=SubscriptionSerializerSchema(),
        responses={(200, "application/json"): SubscriptionSerializerSchema()},
    ),
    destroy=extend_schema(responses={204: None}),
)
class SubscriptionViewSet(
    BaseFHIRView, ListModelMixin, RetrieveModelMixin, ModelViewSet
):
    _error_while_deleting = "Error while deleting a subscription: %(msg)s"
    serializer_class = SubscriptionSerializer
    http_method_names = ("get", "post", "put", "delete")
    permission_classes = (FHIRApiSubscriptionPermissions,)

    def get_queryset(self):
        queryset = Subscription.objects.filter(is_deleted=False).order_by(
            "date_created"
        )
        return DateUpdatedRequestParameterFilter(self.request).filter_queryset(queryset)

    def perform_destroy(self, instance):
        if not self.check_if_owner(self.request.user, instance):
            raise PermissionDenied(
                detail=self._error_while_deleting
                % {"msg": "You are not the owner of this subscription"}
            )
        service = SubscriptionService(self.request.user)
        result = service.delete({"id": instance.uuid})
        self.check_error_message(result)

    @staticmethod
    def check_error_message(result):
        if not result.get("success", False):
            msg = result.get("message", "Unknown")
            raise APIException(f"Error while deleting a subscription: {msg}")

    @staticmethod
    def check_if_owner(user, instance):
        return str(user.id).lower() == str(instance.user_created.id).lower()



from api_fhir_r4.converters.locationConverter import LocationConverter
from fhir.resources.R4B.location import Location
from api_fhir_r4.tests import LocationTestMixin
from api_fhir_r4.tests.mixin import (
    ConvertToImisTestMixin,
    ConvertToFhirTestMixin,
    ConvertJsonToFhirTestMixin,
)


class LocationConverterTestCase(
    LocationTestMixin,
    ConvertToImisTestMixin,
    ConvertToFhirTestMixin,
    ConvertJsonToFhirTestMixin,
):
    converter = LocationConverter
    fhir_resource = Location
    json_repr = "test/test_location.json"



# flake8: noqa

from api_fhir_r4.configurations.baseApiFHIRConfiguration import BaseConfiguration
from api_fhir_r4.configurations.baseApiFHIRConfiguration import IdentifierConfiguration
from api_fhir_r4.configurations.baseApiFHIRConfiguration import LocationConfiguration
from api_fhir_r4.configurations.baseApiFHIRConfiguration import MaritalConfiguration
from api_fhir_r4.configurations.baseApiFHIRConfiguration import IssueTypeConfiguration
from api_fhir_r4.configurations.baseApiFHIRConfiguration import ClaimConfiguration
from api_fhir_r4.configurations.baseApiFHIRConfiguration import CoverageEligibilityConfiguration
from api_fhir_r4.configurations.baseApiFHIRConfiguration import CommunicationRequestConfiguration
from api_fhir_r4.configurations.baseApiFHIRConfiguration import CoverageConfiguration
from api_fhir_r4.configurations.baseApiFHIRConfiguration import InvoiceConfiguration
from api_fhir_r4.configurations.baseApiFHIRConfiguration import OrganisationConfiguration
from api_fhir_r4.configurations.baseApiFHIRConfiguration import SubscriptionConfiguration
from api_fhir_r4.configurations.baseApiFHIRConfiguration import PaymentNoticeConfiguration
from api_fhir_r4.configurations.baseApiFHIRConfiguration import BaseApiFhirConfiguration

from api_fhir_r4.configurations.generalConfiguration import GeneralConfiguration
from api_fhir_r4.configurations.R4IdentifierConfig import R4IdentifierConfig
from api_fhir_r4.configurations.R4LocationConfig import R4LocationConfig
from api_fhir_r4.configurations.R4MaritalConfig import R4MaritalConfig
from api_fhir_r4.configurations.R4IssueTypeConfig import R4IssueTypeConfig
from api_fhir_r4.configurations.R4ClaimConfig import R4ClaimConfig
from api_fhir_r4.configurations.R4InvoiceConfig import R4InvoiceConfig
from api_fhir_r4.configurations.R4CoverageEligibilityConfiguration import (
    R4CoverageEligibilityConfiguration,
)
from api_fhir_r4.configurations.R4CommunicationRequestConfig import (
    R4CommunicationRequestConfig,
)
from api_fhir_r4.configurations.R4OrganisationConfig import R4OrganisationConfig
from api_fhir_r4.configurations.R4CoverageConfig import R4CoverageConfig
from api_fhir_r4.configurations.R4SubscriptionConfig import R4SubscriptionConfig
from api_fhir_r4.configurations.R4PaymentNoticeConfig import R4PaymentNoticeConfig

# all specific configurations have to be imported before R4ApiFhirConfig
from api_fhir_r4.configurations.R4ApiFhirConfig import R4ApiFhirConfig
from api_fhir_r4.configurations.moduleConfiguration import ModuleConfiguration



from rest_framework import viewsets

from api_fhir_r4.mixins import MultiIdentifierRetrieverMixin, MultiIdentifierUpdateMixin
from api_fhir_r4.model_retrievers import (
    UUIDIdentifierModelRetriever,
    CodeIdentifierModelRetriever,
)
from api_fhir_r4.permissions import FHIRApiLocationPermissions
from api_fhir_r4.serializers import LocationSerializer, LocationSiteSerializer
from api_fhir_r4.views.fhir.base import BaseFHIRView
from api_fhir_r4.views.filters import ValidityFromRequestParameterFilter
from location.models import HealthFacility, Location


class LocationViewSet(
    BaseFHIRView,
    MultiIdentifierRetrieverMixin,
    viewsets.ModelViewSet,
    MultiIdentifierUpdateMixin,
):
    retrievers = [UUIDIdentifierModelRetriever, CodeIdentifierModelRetriever]
    serializer_class = LocationSerializer
    permission_classes = (FHIRApiLocationPermissions,)

    def list(self, request, *args, **kwargs):
        identifier = request.GET.get("identifier")
        physical_type = request.GET.get("physicalType")
        queryset = self.get_queryset(physical_type)
        if identifier:
            return self.retrieve(request, *args, **{**kwargs, "identifier": identifier})
        else:
            queryset = queryset.filter(validity_to__isnull=True).order_by(
                "validity_from"
            )
        if physical_type and physical_type == "si":
            self.serializer_class = LocationSiteSerializer
            serializer = LocationSiteSerializer(
                self.paginate_queryset(queryset), many=True, user=request.user
            )
        else:
            serializer = LocationSerializer(
                self.paginate_queryset(queryset), many=True, user=request.user
            )
        return self.get_paginated_response(serializer.data)

    def retrieve(self, *args, **kwargs):
        physical_type = self.request.GET.get("physicalType")
        if physical_type and physical_type == "si":
            self.serializer_class = LocationSiteSerializer
            self.queryset = self.get_queryset("si")
        response = super().retrieve(self, *args, **kwargs)
        return response

    def get_queryset(self, physicalType="area"):
        if physicalType == "si":
            hf_queryset = HealthFacility.get_queryset(None, self.request.user)
            queryset = (
                hf_queryset.select_related("location")
                .select_related("sub_level")
                .select_related("legal_form")
            )
        else:
            queryset = Location.get_queryset(None, self.request.user)
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(
            queryset
        )



# Generated by Django 3.0.14 on 2022-03-13 16:07

import core.datetimes.ad_datetime
import core.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("api_fhir_r4", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="historicalsubscription",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical subscription",
            },
        ),
        migrations.CreateModel(
            name="SubscriptionNotificationResult",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        db_column="UUID",
                        default=None,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("notified_successfully", models.BooleanField()),
                (
                    "notification_time",
                    core.fields.DateTimeField(
                        db_column="Expiring",
                        default=core.datetimes.ad_datetime.AdDatetime.now,
                    ),
                ),
                ("error", models.TextField(default=None, null=True)),
                (
                    "subscription",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="notifications_sent",
                        to="api_fhir_r4.Subscription",
                    ),
                ),
            ],
            options={
                "db_table": "tblSubscriptionNotificationResult",
                "managed": True,
            },
        ),
    ]



# Generated by Django 3.0.14 on 2022-01-24 15:20

import core.fields
import datetime
import dirtyfields.dirtyfields
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django_cryptography.fields
import simple_history.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Subscription",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        db_column="UUID",
                        default=None,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "is_deleted",
                    models.BooleanField(db_column="isDeleted", default=False),
                ),
                (
                    "json_ext",
                    models.JSONField(blank=True, db_column="Json_ext", null=True),
                ),
                (
                    "date_created",
                    core.fields.DateTimeField(db_column="DateCreated", null=True),
                ),
                (
                    "date_updated",
                    core.fields.DateTimeField(db_column="DateUpdated", null=True),
                ),
                ("version", models.IntegerField(default=1)),
                (
                    "date_valid_from",
                    core.fields.DateTimeField(
                        db_column="DateValidFrom", default=datetime.datetime.now
                    ),
                ),
                (
                    "date_valid_to",
                    core.fields.DateTimeField(
                        blank=True, db_column="DateValidTo", null=True
                    ),
                ),
                (
                    "replacement_uuid",
                    models.UUIDField(db_column="ReplacementUUID", null=True),
                ),
                (
                    "status",
                    models.SmallIntegerField(
                        choices=[(0, "inactive"), (1, "active")], db_column="Status"
                    ),
                ),
                (
                    "channel",
                    models.SmallIntegerField(
                        choices=[(0, "rest-hook")], db_column="Channel"
                    ),
                ),
                ("endpoint", models.CharField(db_column="Endpoint", max_length=255)),
                (
                    "headers",
                    django_cryptography.fields.encrypt(
                        models.TextField(db_column="Headers", max_length=255, null=True)
                    ),
                ),
                ("criteria", models.JSONField(db_column="Criteria", null=True)),
                ("expiring", models.DateTimeField(db_column="Expiring")),
                (
                    "user_created",
                    models.ForeignKey(
                        db_column="UserCreatedUUID",
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="subscription_user_created",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "user_updated",
                    models.ForeignKey(
                        db_column="UserUpdatedUUID",
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="subscription_user_updated",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "db_table": "tblSubscription",
                "managed": True,
            },
            bases=(dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalSubscription",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        db_column="UUID", db_index=True, default=None, editable=False
                    ),
                ),
                (
                    "is_deleted",
                    models.BooleanField(db_column="isDeleted", default=False),
                ),
                (
                    "json_ext",
                    models.JSONField(blank=True, db_column="Json_ext", null=True),
                ),
                (
                    "date_created",
                    core.fields.DateTimeField(db_column="DateCreated", null=True),
                ),
                (
                    "date_updated",
                    core.fields.DateTimeField(db_column="DateUpdated", null=True),
                ),
                ("version", models.IntegerField(default=1)),
                (
                    "date_valid_from",
                    core.fields.DateTimeField(
                        db_column="DateValidFrom", default=datetime.datetime.now
                    ),
                ),
                (
                    "date_valid_to",
                    core.fields.DateTimeField(
                        blank=True, db_column="DateValidTo", null=True
                    ),
                ),
                (
                    "replacement_uuid",
                    models.UUIDField(db_column="ReplacementUUID", null=True),
                ),
                (
                    "status",
                    models.SmallIntegerField(
                        choices=[(0, "inactive"), (1, "active")], db_column="Status"
                    ),
                ),
                (
                    "channel",
                    models.SmallIntegerField(
                        choices=[(0, "rest-hook")], db_column="Channel"
                    ),
                ),
                ("endpoint", models.CharField(db_column="Endpoint", max_length=255)),
                (
                    "headers",
                    django_cryptography.fields.encrypt(
                        models.TextField(db_column="Headers", max_length=255, null=True)
                    ),
                ),
                ("criteria", models.JSONField(db_column="Criteria", null=True)),
                ("expiring", models.DateTimeField(db_column="Expiring")),
                ("history_id", models.AutoField(primary_key=True, serialize=False)),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "user_created",
                    models.ForeignKey(
                        blank=True,
                        db_column="UserCreatedUUID",
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "user_updated",
                    models.ForeignKey(
                        blank=True,
                        db_column="UserUpdatedUUID",
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical subscription",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
    ]



import logging
from typing import List

from django.core.exceptions import MultipleObjectsReturned
from django.db import models, IntegrityError

from api_fhir_r4.containedResources.converters import (
    FHIRContainedResourceConverter,
    IMISContainedResourceConverter,
)
from fhir.resources.R4B import FHIRAbstractModel

from api_fhir_r4.serializers import BaseFHIRSerializer

logger = logging.getLogger(__name__)


class ContainedResourceManager:
    """It's used for managing contained resources in fhir objects

    Methods
    ----------
    convert_to_fhir(imis_obj: models.Model):
        Convert IMIS Model attribute to FHIR Object

    convert_to_imis(imis_obj: models.Model):
        Convert FHIR Object contained resource to IMIS Model

    create_or_update_from_contained(self, fhir_model: dict) -> List[models.Model]
        Takes contained resources from fhir dict representation and saves them in database. If resource
        for given resource ID already exists then it updates object definition.
    """

    def __init__(
        self,
        fhir_converter: FHIRContainedResourceConverter = None,
        imis_converter: IMISContainedResourceConverter = None,
        serializer: BaseFHIRSerializer = None,
        alias: str = None,
    ):
        self.fhir_converter = fhir_converter
        self.imis_converter = imis_converter
        self.serializer = serializer
        self.alias = alias

    def convert_to_fhir(self, imis_obj: models.Model) -> List[FHIRAbstractModel]:
        self._assert_fhir_converter()
        return self.fhir_converter.convert(imis_obj)

    def convert_to_imis(self, fhir_model: dict) -> List[models.Model]:
        self._assert_imis_converter()
        return self.imis_converter.convert(
            fhir_model, self.serializer.get_audit_user_id()
        )

    def create_or_update_from_contained(self, fhir_model: dict) -> List[models.Model]:
        self._assert_serializer()
        imis_representation = self.convert_to_imis(fhir_model)
        output = []
        for instance in imis_representation:
            if instance.uuid and self._is_saved_in_db(instance):
                output.append(self._update(instance))
            else:
                output.append(self._create(instance))

        return output

    def _assert_fhir_converter(self):
        assert self.fhir_converter is not None, "FHIR Converter is required"

    def _assert_imis_converter(self):
        assert self.imis_converter is not None, "IMIS Converter is required"

    def _assert_serializer(self):
        assert (
            self.serializer is not None
        ), "Serializer is required to perform create and update"

    def _update(self, updated_instance):
        try:
            instance = updated_instance.__class__.objects.get(
                uuid=updated_instance.uuid
            )
        except MultipleObjectsReturned as a:
            logger.error(a)
            raise IntegrityError(
                f"While trying to use resource {updated_instance} for update of object"
                f" with uuid {updated_instance.uuid} multiple objects with this uuid were found"
            ) from a

        try:
            updated = self.serializer.update(
                instance, self._model_to_dict(updated_instance)
            )
            updated.save()
            return updated
        except Exception as e:
            import warnings

            warnings.warn(
                f"Update from contained resource failed due to error: \n{e}.\n"
                f"Instance will not be updated and default value will be used"
            )
            return instance

    def _create(self, instance):
        # Services from other modules are often called through update_or_create method. It treats objects with uuid
        # as updatable. If non-existing object with explicitly given uuid is about to be created it'll try to update
        # it instead. Therefore, uuid is temporary removed and overwritten after object is created.
        as_dict = self._model_to_dict(instance)
        uuid = as_dict.get("uuid", None)
        as_dict["uuid"] = None
        created = self.serializer.create(as_dict)
        if uuid:
            created.uuid = uuid
            created.save()
        return created

    def _is_saved_in_db(self, obj: models.Model):
        # Checks if given ID is already stored in database
        return obj.__class__.objects.filter(uuid=obj.uuid).exists()

    def _model_to_dict(self, instance):
        # Due to how serializers are build simple __dict__ is used instead of builtin model_to_dict
        return {k: v for k, v in instance.__dict__.items() if not k.startswith("_")}




from medical.models import Item

from api_fhir_r4.converters import MedicationConverter
from api_fhir_r4.serializers.baseSerializer import BaseFHIRSerializer


class MedicationSerializer(BaseFHIRSerializer):
    fhirConverter = MedicationConverter

    def create(self, validated_data):
        imis_medication = Item(**validated_data)
        imis_medication.audit_user_id = self.get_audit_user_id()
        filters = {}
        if imis_medication.id:
            filters["id"] = imis_medication.id
        if imis_medication.uuid:
            filters["uuid"] = imis_medication.uuid
        if imis_medication.code and not filters:
            filters["code"] = imis_medication.code
            filters["validity_to__isnull"] = True

        instance = Item.objects.filter(**filters).first()
        if instance:
            raise ValueError(
                f"cannot create an already existing object, filters: {filters}"
            )
        imis_medication.audit_user_id = self.get_audit_user_id()
        imis_medication.save()
        return imis_medication

    def update(self, instance, validated_data):
        imis_medication = Item(**validated_data)
        imis_medication.audit_user_id = self.get_audit_user_id()
        filters = {}
        if imis_medication.id:
            filters["id"] = imis_medication.id
        if imis_medication.uuid:
            filters["uuid"] = imis_medication.uuid
        if imis_medication.code and not filters:
            filters["code"] = imis_medication.code
            filters["validity_to__isnull"] = True

        instance = Item.objects.filter(**filters).first()
        if not instance:
            raise ValueError(f"cannot update a not found object, filters: {filters}")
        instance.audit_user_id = self.get_audit_user_id()
        instance.save_history()
        imis_medication.id = instance.id
        imis_medication.uuid = instance.uuid
        imis_medication.code = instance.code
        imis_medication.save()
        return imis_medication



from claim.services import ClaimSubmitService, ClaimConfig
from claim.gql_mutations import create_attachments
from claim.models import Claim, ClaimItem, ClaimService
from typing import List, Union
from api_fhir_r4.containedResources.claimContainedResources import (
    ClaimContainedResources,
)
from api_fhir_r4.containedResources.serializerMixin import (
    ContainedContentSerializerMixin,
)
from api_fhir_r4.models import ClaimV2 as FHIRClaim
from django.http import HttpResponseForbidden
from django.http.response import HttpResponseBase
from django.shortcuts import get_object_or_404

from api_fhir_r4.configurations import R4ClaimConfig, GeneralConfiguration
from api_fhir_r4.converters import (
    ClaimResponseConverter,
    OperationOutcomeConverter,
    ReferenceConverterMixin as r,
)
from api_fhir_r4.converters.claimConverter import ClaimConverter
from fhir.resources.R4B import FHIRAbstractModel
from api_fhir_r4.serializers.baseSerializer import BaseFHIRSerializer
import logging

logger = logging.getLogger(__name__)


class ClaimSerializer(ContainedContentSerializerMixin, BaseFHIRSerializer):

    fhirConverter = ClaimConverter

    contained_resources = ClaimContainedResources

    def fhir_object_reference_fields(
        self, fhir_obj: FHIRClaim
    ) -> List[FHIRAbstractModel]:
        return [
            fhir_obj.patient,
            fhir_obj.provider,
            fhir_obj.enterer,
            *[item.extension[0].valueReference for item in fhir_obj.item],
        ]

    def create(self, validated_data):
        from_contained = self._create_or_update_contained(self.initial_data)
        claim = self._create_claim_from_validated_data(validated_data, from_contained)
        return self.create_claim_response(claim.code)

    def create_claim_response(self, claim_code):
        claim = get_object_or_404(Claim, code=claim_code, *Claim.filter_validity())
        return ClaimResponseConverter.to_fhir_obj(claim)

    def create_claim_attachments(self, claim_code, attachments):
        claim = get_object_or_404(Claim, code=claim_code, *Claim.filter_validity())
        create_attachments(claim.id, attachments)

    def to_representation(self, obj):
        if isinstance(obj, HttpResponseBase):
            return OperationOutcomeConverter.to_fhir_obj(obj).dict()
        elif isinstance(obj, FHIRAbstractModel):
            return obj.dict()

        fhir_obj = self.fhirConverter.to_fhir_obj(obj, self._reference_type)
        self.remove_attachment_data(fhir_obj)

        if self.context.get("contained", None):
            self._add_contained_references(fhir_obj)

        fhir_dict = fhir_obj.dict()
        if self.context.get("contained", False):
            fhir_dict["contained"] = self._create_contained_obj_dict(obj)
        return fhir_dict

    def remove_attachment_data(self, fhir_obj):
        if hasattr(self.parent, "many") and self.parent.many is True:
            attachments = self.__get_attachments(fhir_obj)
            for next_attachment in attachments:
                next_attachment.data = None

    @property
    def reference_type(self):
        return super().reference_type

    @reference_type.setter
    def reference_type(
        self,
        reference_type: Union[
            r.UUID_REFERENCE_TYPE, r.CODE_REFERENCE_TYPE, r.DB_ID_REFERENCE_TYPE
        ],
    ):
        if reference_type != self._reference_type:
            self._reference_type = reference_type
            self.__set_contained_resource_reference_types(reference_type)

    def __get_attachments(self, fhir_obj):
        attachment_category = R4ClaimConfig.get_fhir_claim_attachment_code()
        return [
            a.valueAttachment
            for a in fhir_obj.supportingInfo
            if a.category.text == attachment_category
        ]

    def __set_contained_resource_reference_types(self, reference_type):
        self._contained_definitions.update_reference_type(reference_type)

    def __claim_provisions_to_dict(self, list_of_provisions, contained_items):
        # Claim Entering service is expecting to receive items and services in form of
        # dict and then uses process_items_relations or process_services_relations to create actual items.
        out = []
        for x in list_of_provisions:
            dict_ = x.__dict__
            dict_.pop("_state", None)
            if isinstance(x, ClaimItem):
                dict_["item_id"] = (
                    self.__get_contained_medical_provision(contained_items, dict_)
                    or dict_["item_id"]
                )
            elif isinstance(x, ClaimService):
                dict_["service_id"] = (
                    self.__get_contained_medical_provision(contained_items, dict_)
                    or dict_["service_id"]
                )
            else:
                raise AttributeError(
                    f"Medical provision {x} is not ClaimItem nor ClaimService"
                )

            dict_.pop("code")
            out.append(dict_)
        return out

    def __get_contained_or_default_hf_id(self, contained_dict, validated_data):
        return self.__get_contained_or_default(
            contained_dict,
            validated_data,
            "health_facility__Organization",
            "health_facility_code",
            "health_facility_id",
        )

    def __get_contained_or_default(
        self, contained_dict, validated_data, ref_attr, ref_code, ref_id, code="code"
    ):
        c_elm = contained_dict.pop(ref_attr, None)
        elm_id = None
        if c_elm and ref_code in validated_data:
            elm_id = self.__id_from_contained(
                c_elm,
                lambda x: hasattr(x, code)
                and getattr(x, code) == validated_data[ref_code],
            )
        if elm_id is None and ref_id in validated_data:
            elm_id = validated_data[ref_id]
        return elm_id

    def __get_contained_or_default_insuree(self, contained_dict, validated_data):
        return self.__get_contained_or_default(
            contained_dict,
            validated_data,
            "insuree__Patient",
            "insuree_chf_id",
            "insuree_id",
            "chf_id",
        )

    def __get_contained_or_default_claim_admin(self, contained_dict, validated_data):
        return self.__get_contained_or_default(
            contained_dict,
            validated_data,
            "admin__Practitioner",
            "claim_admin_code",
            "admin_id",
        )

    def __get_contained_medical_provision(self, contained_items: list, item):
        contained_value = self.__id_from_contained(
            contained_items, lambda x: hasattr(x, "code") and x.code == item["code"]
        )
        return contained_value

    def __id_from_contained(self, contained_collection, lookup_func):
        matching = [x.id for x in contained_collection if lookup_func(x)]
        return matching[0] if len(matching) != 0 else None

    def _create_claim_from_validated_data(self, validated_data, contained):
        truncated_data = self._claim_input_from_validated_claim_data(
            validated_data, contained
        )
        user = self.context.get("request").user
        if not user or not user.has_perms(
            ClaimConfig.gql_mutation_create_claims_perms
            + ClaimConfig.gql_mutation_submit_claims_perms
        ):
            return HttpResponseForbidden()

        rule_engine_validation = GeneralConfiguration.get_claim_rule_engine_validation()
        claim = ClaimSubmitService(user).enter_and_submit(
            truncated_data, rule_engine_validation=rule_engine_validation
        )

        attachments = validated_data.get("claim_attachments", [])
        self.create_claim_attachments(claim.code, attachments)
        return claim

    def _claim_input_from_validated_claim_data(self, validated_data, contained):
        essential_claim_fields = [
            "date_claimed",
            "date_from",
            "date_to",
            "icd_id",
            "icd_1_id",
            "icd_2_id",
            "icd_3_id",
            "icd_4_id",
            "code",
            "claimed",
            "explanation",
            "adjustment",
            "category",
            "visit_type",
            "guarantee_id",
            "insuree_id",
            "health_facility_id",
            "admin_id",
            "items",
            "services",
            "json_ext",
        ]
        truncated_data = {
            k: v for k, v in validated_data.items() if k in essential_claim_fields
        }
        truncated_data.pop("_state", None)

        # Items and services passed in converter through additional attributes
        # TODO : check that items__Medication could be a key of contained
        truncated_data["items"] = self.__claim_provisions_to_dict(
            validated_data["submit_items"], contained.get("items__Medication", [])
        )

        truncated_data["services"] = self.__claim_provisions_to_dict(
            validated_data["submit_services"],
            contained.get("services__ActivityDefinition", []),
        )

        # If those resources are created through contained id is not available until _create_or_update_contained is
        # called. This ensures references are ok. Codes are assigned in converter as additional variables.
        # .get() is used as those values are mandatory in claim and have to be unique.
        truncated_data["health_facility_id"] = self.__get_contained_or_default_hf_id(
            contained, validated_data
        )
        truncated_data["insuree_id"] = self.__get_contained_or_default_insuree(
            contained, validated_data
        )
        truncated_data["admin_id"] = self.__get_contained_or_default_claim_admin(
            contained, validated_data
        )
        return truncated_data



import logging
from collections import defaultdict, OrderedDict
from datetime import datetime as py_datetime
from django.db.models import Q
from django.http import Http404
from itertools import chain
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework import status

from api_fhir_r4.defaultConfig import DEFAULT_CFG
from api_fhir_r4.mixins import (
    MultiIdentifierRetrieveManySerializersMixin,
    MultiIdentifierRetrieverMixin,
)
from api_fhir_r4.model_retrievers import (
    CodeIdentifierModelRetriever,
    DatabaseIdentifierModelRetriever,
    UUIDIdentifierModelRetriever,
)
from api_fhir_r4.multiserializer import modelViewset
from api_fhir_r4.permissions import (
    FHIRApiOrganizationPermissions,
    FHIRApiHealthServicePermissions,
    FHIRApiInsuranceOrganizationPermissions,
)
from api_fhir_r4.views.fhir.base import BaseMultiserializerFHIRView
from api_fhir_r4.serializers import (
    PolicyHolderOrganisationSerializer,
    HealthFacilityOrganisationSerializer,
    InsuranceOrganizationSerializer,
)
from api_fhir_r4.views.filters import (
    ValidityFromRequestParameterFilter,
    DateUpdatedRequestParameterFilter,
)
from location.models import HealthFacility
from core.models import ModuleConfiguration
from policyholder.models import PolicyHolder

logger = logging.getLogger(__name__)


class OrganisationViewSet(
    BaseMultiserializerFHIRView,
    modelViewset.MultiSerializerModelViewSet,
    MultiIdentifierRetrieveManySerializersMixin,
    MultiIdentifierRetrieverMixin,
):
    retrievers = [
        UUIDIdentifierModelRetriever,
        DatabaseIdentifierModelRetriever,
        CodeIdentifierModelRetriever,
    ]

    lookup_field = "identifier"

    @property
    def serializers(self):
        return {
            HealthFacilityOrganisationSerializer: (
                self._hf_queryset(),
                self._hf_serializer_validator,
                (FHIRApiHealthServicePermissions,),
            ),
            PolicyHolderOrganisationSerializer: (
                self._ph_queryset(),
                self._ph_serializer_validator,
                (FHIRApiOrganizationPermissions,),
            ),
            InsuranceOrganizationSerializer: (
                self._io_queryset(),
                self._io_serializer_validator,
                (FHIRApiInsuranceOrganizationPermissions,),
            ),
        }

    @classmethod
    def _hf_serializer_validator(cls, context):
        return cls._base_request_validator_dispatcher(
            request=context["request"],
            get_check=lambda x: cls._get_type_from_query(x) in ("prov", None),
            post_check=lambda x: cls._get_type_from_body(x) == "prov",
            put_check=lambda x: cls._get_type_from_body(x) in ("prov", None),
        )

    @classmethod
    def _ph_serializer_validator(cls, context):
        return cls._base_request_validator_dispatcher(
            request=context["request"],
            get_check=lambda x: cls._get_type_from_query(x) in ("bus", None),
            post_check=lambda x: cls._get_type_from_body(x) == "bus",
            put_check=lambda x: cls._get_type_from_body(x) in ("bus", None),
        )

    @classmethod
    def _io_serializer_validator(cls, context):
        return cls._base_request_validator_dispatcher(
            request=context["request"],
            get_check=lambda x: cls._get_type_from_query(x) in ("ins", None),
            post_check=lambda x: cls._get_type_from_body(x) == "ins",
            put_check=lambda x: cls._get_type_from_body(x) in ("ins", None),
        )

    @classmethod
    def _base_request_validator_dispatcher(
        cls, request: Request, get_check, post_check, put_check
    ):
        if request.method == "GET":
            return get_check(request)
        elif request.method == "POST":
            return post_check(request)
        elif request.method == "PUT":
            return put_check(request)
        return True

    def list(self, request, *args, **kwargs):
        resource_type = request.GET.get("resourceType")
        identifier = request.GET.get("code")
        if identifier:
            return self.retrieve(request, *args, **{**kwargs, "identifier": identifier})

        self._validate_list_model_request()
        filtered_querysets = {}  # {serialzer: qs}

        for serializer, (qs, _, _) in self.get_eligible_serializers_iterator():
            next_serializer_data = self.filter_queryset(qs)
            model = next_serializer_data.model
            filtered_querysets[model, serializer] = next_serializer_data

        # if insurance organisation queryset is empty - take the default one
        if resource_type is None or resource_type == "ins":
            if (
                ModuleConfiguration,
                InsuranceOrganizationSerializer,
            ) in filtered_querysets:
                if (
                    len(
                        filtered_querysets[
                            ModuleConfiguration, InsuranceOrganizationSerializer
                        ]
                    )
                    == 0
                ):
                    filtered_querysets[
                        ModuleConfiguration, InsuranceOrganizationSerializer
                    ] = [DEFAULT_CFG["R4_fhir_insurance_organisation_config"]]
                else:
                    # save in 'filtered_queryset' values from module db config to have good value in 'total_count'
                    filtered_querysets[
                        ModuleConfiguration, InsuranceOrganizationSerializer
                    ] = self._get_insurance_organisations_as_list()

        page = self.paginate_queryset(list(chain(*filtered_querysets.values())))
        data = self.__dispatch_page_data(page)
        serialized_data = self._serialize_dispatched_data(
            data, dict(filtered_querysets.keys()), user=request.user
        )
        data = self.get_paginated_response(serialized_data)
        return data

    def __dispatch_page_data(self, page):
        x = defaultdict(list)
        for r in page:
            x[type(r)].append(r)
        return x

    def retrieve(self, request, *args, **kwargs):
        self._validate_retrieve_model_request()
        retrieved = []
        errors = []
        for serializer, (qs, _, _) in self.get_eligible_serializers_iterator():
            if qs.model is not ModuleConfiguration:
                try:
                    ref_type, instance = self._get_object_with_first_valid_retriever(
                        qs, kwargs["identifier"]
                    )
                    if instance:
                        serializer = serializer(
                            instance, reference_type=ref_type, user=request.user
                        )
                        if serializer.data:
                            retrieved.append(serializer.data)
                except Exception as e:
                    # we will try with another retriever
                    errors.append(str(e))
            else:
                if qs.count() > 0:
                    data = self._get_insurance_organisation(
                        kwargs.get("identifier", None)
                    )
                else:
                    data = self._get_insurance_organisation_default(
                        kwargs.get("identifier", None)
                    )
                if data:
                    retrieved.append(data)

        if len(retrieved) > 1:
            raise ValueError(
                "Ambiguous retrieve result, object found for multiple serializers."
            )
        elif len(retrieved) == 0:
            if len(errors) > 0:
                raise Http404(
                    f"Resource for identifier {kwargs['identifier']} cannot be reloved:{','.join(errors)}"
                )
            else:
                raise Http404(
                    f"Resource for identifier {kwargs['identifier']} not found"
                )

        return Response(retrieved[0])

    def _get_insurance_organisations_as_list(self):
        data = []
        module_config = self._io_queryset()[0]._cfg
        if "insurer_organisation" in module_config:
            insurer_organisations = module_config["insurer_organisation"]
            for ins in insurer_organisations:
                data.append(ins)
        return data

    def _get_insurance_organisations(self):
        """method to get insurance organisation from module fhir config"""
        data = []
        serializer = InsuranceOrganizationSerializer()
        module_config = self._io_queryset()[0]._cfg
        if "insurer_organisation" in module_config:
            insurer_organisations = module_config["insurer_organisation"]
            for ins in insurer_organisations:
                data.append(serializer.to_representation(obj=ins))
        return data

    def _get_insurance_organisation(self, identifier):
        """method to get chosen insurance organisation from module fhir config"""
        insurer_organisations = self._get_insurance_organisations()
        if len(insurer_organisations) > 0:
            for ins in insurer_organisations:
                if identifier == f"{ins['id']}":
                    return ins

    def _get_insurance_organisation_default(self, identifier, user=None):
        """method to get chosen insurance organisation from default config if no config in db"""
        serializer = InsuranceOrganizationSerializer(user=user)
        default_insurance_organisation = DEFAULT_CFG[
            "R4_fhir_insurance_organisation_config"
        ]
        if identifier == f"{default_insurance_organisation['id']}":
            return serializer.to_representation(obj=default_insurance_organisation)

    def get_queryset(self):
        return HealthFacility.objects

    def _hf_queryset(self):
        queryset = HealthFacility.objects.filter(validity_to__isnull=True).order_by(
            "validity_from"
        )
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(
            queryset
        )

    def _ph_queryset(self):
        queryset = PolicyHolder.objects.filter(is_deleted=False).order_by(
            "date_created"
        )
        return DateUpdatedRequestParameterFilter(self.request).filter_queryset(queryset)

    def _io_queryset(self):
        now = py_datetime.now()  # can't use core config here...
        queryset = ModuleConfiguration.objects.filter(
            Q(is_disabled_until=None) | Q(is_disabled_until__lt=now),
            layer="be",
            module="api_fhir_r4",
        )
        return DateUpdatedRequestParameterFilter(self.request).filter_queryset(queryset)

    @classmethod
    def _get_type_from_body(cls, request):
        try:
            # See: http://hl7.org/fhir/R4/organization.html
            return request.data["type"][0]["coding"][0]["code"].lower()
        except KeyError:
            return None

    @classmethod
    def _get_type_from_query(cls, request):
        try:
            return request.GET["type"].lower()
        except KeyError:
            return None

    def _serialize_dispatched_data(self, data, serializer_models, user=None):
        # override this method to support InsuranceOrganisation serializer
        #  solution - model taken from ModuleConfiguration model
        serialized = []
        for model, model_data in data.items():
            serializer_cls = serializer_models.get(model, None)
            if not serializer_cls:
                if all(isinstance(md, OrderedDict) for md in model_data):
                    serialized.extend(self._get_insurance_organisations())
                else:
                    # check if we have insurance organisation default config
                    data_default = self.__check_default_config_insurance_organisation(
                        model_data, user=user
                    )
                    if len(data_default) > 0:
                        serialized.extend(data_default)
                    else:
                        logger.error(
                            f"Found data of type {model_data} but it couldn't be matched with "
                            f"any of available serializers {serializer_models}"
                        )
                    continue
            else:
                serializer = serializer_cls(tuple(model_data), many=True, user=user)
                serialized.extend(serializer.data)

        return serialized

    def __check_default_config_insurance_organisation(self, model_data, user=None):
        for md in model_data:
            if md.get("resource_type") == "insurance_organisation":
                identifier = DEFAULT_CFG["R4_fhir_insurance_organisation_config"]["id"]
                return [
                    self._get_insurance_organisation_default(
                        identifier=identifier, user=user
                    )
                ]

    def create(self, request, *args, **kwargs):
        """
        Override to ensure MultiSerializerCreateModelMixin.create() is used instead of
        BaseFHIRView's CreateModelMixin.create(). The base CreateModelMixin calls
        get_serializer() which returns MultiSerializerSerializerClass (a placeholder
        that doesn't accept data), causing the "no data=" error.
        """
        self._validate_create_request()
        results = []
        for serializer, _ in self.get_eligible_serializers_iterator():
            data = self._create_for_serializer(serializer, request, *args, **kwargs)
            results.append(data)

        headers = self.get_success_headers(results)
        response = results[0]  # By default there should be only one eligible serializer
        return Response(response, status=status.HTTP_201_CREATED, headers=headers)



from abc import ABC

from api_fhir_r4.multiserializer import mixins
from rest_framework.viewsets import GenericViewSet

from api_fhir_r4.multiserializer.serializerClass import MultiSerializerSerializerClass


class MultiSerializerModelViewSet(
    GenericViewSet,
    mixins.MultiSerializerCreateModelMixin,
    mixins.MultiSerializerRetrieveModelMixin,
    mixins.MultiSerializerUpdateModelMixin,
    mixins.MultiSerializerListModelMixin,
    ABC,
):
    serializer_class = MultiSerializerSerializerClass



from django.contrib.contenttypes.models import ContentType
from api_fhir_r4.converters import CodeSystemConverter
from api_fhir_r4.serializers.baseSerializer import BaseFHIRSerializer


class CodeSystemSerializer(BaseFHIRSerializer):
    fhirConverter = CodeSystemConverter
    codeSystemFields = [
        "code_field",
        "display_field",
        "id",
        "name",
        "title",
        "description",
        "url",
    ]

    def __init__(self, *args, **kwargs):
        self.model = {}
        if "user" in kwargs:
            user = kwargs.pop("user")
        for field in self.codeSystemFields:
            self.model[field] = kwargs.pop(field, None)
        if "data" in kwargs:
            self.model["data"] = kwargs.pop("data")
        elif "model_name" in kwargs:
            content_type = ContentType.objects.get(
                model__iexact=kwargs.pop("model_name")
            )
            model_class = content_type.model_class()
            self.model["data"] = model_class.objects.all()
        else:
            self.model["data"] = {}

        super().__init__(*args, user=user, **kwargs)

    def to_representation(self, obj):
        return CodeSystemConverter.to_fhir_obj(self.model, self.reference_type).dict()



import datetime

from rest_framework.viewsets import GenericViewSet
from api_fhir_r4.mixins import ListModelMixin, RetrieveModelMixin
from api_fhir_r4.permissions import FHIRApiCoverageRequestPermissions
from api_fhir_r4.serializers.coverageSerializer import CoverageSerializer
from api_fhir_r4.views.fhir.base import BaseFHIRView
from api_fhir_r4.views.filters import ValidityFromRequestParameterFilter
from policy.models import Policy


class CoverageRequestQuerySet(
    BaseFHIRView, RetrieveModelMixin, ListModelMixin, GenericViewSet
):
    lookup_field = "uuid"
    serializer_class = CoverageSerializer
    permission_classes = (FHIRApiCoverageRequestPermissions,)

    def list(self, request, *args, **kwargs):
        queryset = self.get_queryset()
        queryset.prefetch_related("services")
        refDate = request.GET.get("refDate")
        refEndDate = request.GET.get("refEndDate")
        identifier = request.GET.get("identifier")
        if identifier:
            queryset = queryset.filter(chf_id=identifier)
        else:
            queryset = queryset.filter(validity_to__isnull=True).order_by(
                "validity_from"
            )
            if refDate is not None:
                try:
                    datevar = datetime.datetime.strptime(refDate, "%Y-%m-%d").date()
                except ValueError:
                    pass
                finally:
                    queryset = queryset.filter(validity_from__gte=datevar)
            if refEndDate is not None:
                try:
                    datevar = datetime.datetime.strptime(refEndDate, "%Y-%m-%d").date()
                except ValueError:
                    pass
                finally:
                    queryset = queryset.filter(validity_from__lt=datevar)

        serializer = CoverageSerializer(
            self.paginate_queryset(queryset), many=True, user=request.user
        )
        return self.get_paginated_response(serializer.data)

    def get_queryset(self):
        queryset = Policy.get_queryset(None, self.request.user)
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(
            queryset
        )



import copy
from api_fhir_r4.converters import CommunicationConverter
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.serializers.baseSerializer import BaseFHIRSerializer
from claim.models import Claim, Feedback


class CommunicationSerializer(BaseFHIRSerializer):

    fhirConverter = CommunicationConverter

    def create(self, validated_data):
        claim = validated_data.get("claim_id")
        if Claim.objects.filter(id=claim, validity_to__isnull=True).count() == 0:
            raise FHIRException("Claim does not exist")
        if (
            Feedback.objects.filter(claim__id=claim, validity_to__isnull=True).count()
            > 0
        ):
            raise FHIRException("Feedback exists for this claim")
        copied_data = copy.deepcopy(validated_data)
        if "_state" in copied_data:
            del copied_data["_state"]
        from core import datetime

        copied_data["feedback_date"] = datetime.datetime.now()
        obj = Feedback.objects.create(**copied_data)
        imis_claim = Claim.objects.get(id=claim)
        imis_claim.feedback_status = Claim.FEEDBACK_DELIVERED
        imis_claim.feedback_available = True
        imis_claim.feedback = obj
        imis_claim.save()
        return obj



import logging
from typing import Union

from django.http.response import HttpResponseBase
from fhir.resources.R4B import FHIRAbstractModel
from rest_framework import serializers

from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.converters import (
    BaseFHIRConverter,
    OperationOutcomeConverter,
    ReferenceConverterMixin,
)
from core.models import User, TechnicalUser


logger = logging.getLogger(__name__)


class BaseFHIRSerializer(serializers.Serializer):
    fhirConverter = BaseFHIRConverter
    user = None

    def __init__(self, *args, user=None, **kwargs):
        self._reference_type = kwargs.pop(
            "reference_type", ReferenceConverterMixin.UUID_REFERENCE_TYPE
        )
        if user:
            self.user = user
        else:
            context = kwargs.get("context", None)
            if context and hasattr(context, "user"):
                self.user = context.user
        super().__init__(*args, **kwargs)

    def to_representation(self, obj):
        try:
            if isinstance(obj, HttpResponseBase):
                return OperationOutcomeConverter.to_fhir_obj(obj).dict()
            elif isinstance(obj, FHIRAbstractModel):
                return obj.dict()
            return (
                self.fhirConverter(user=self.user)
                .to_fhir_obj(obj, self.reference_type)
                .dict()
            )
        except Exception as e:
            from django.conf import settings

            if settings.DEBUG:
                self._print_debug_log(e)
            raise e

    def to_internal_value(self, data):
        audit_user_id = self.get_audit_user_id()
        imis_obj = self.fhirConverter(user=self.user).to_imis_obj(data, audit_user_id)
        # Filter out special attributes
        return {
            k: v
            for k, v in imis_obj.__dict__.items()
            if not k.startswith("_") and v is not None
        }

    def create(self, validated_data):
        raise NotImplementedError("`create()` must be implemented.")  # pragma: no cover

    def update(self, instance, validated_data):
        raise NotImplementedError("`update()` must be implemented.")  # pragma: no cover

    def get_audit_user_id(self):
        # the audit user is the user
        if self.user:
            return self.user.id_for_audit or -1
        audit_user_id = GeneralConfiguration.get_default_audit_user_id()
        if isinstance(audit_user_id, int):
            return audit_user_id
        else:
            raise ValueError("User not available from the request for audit trail")
            # return self.__get_technical_audit_user(audit_user_id)

    @property
    def reference_type(self):
        return self._reference_type

    @reference_type.getter
    def reference_type(self):
        return self._reference_type

    @reference_type.setter
    def reference_type(
        self,
        reference_type: Union[
            ReferenceConverterMixin.UUID_REFERENCE_TYPE,
            ReferenceConverterMixin.CODE_REFERENCE_TYPE,
            ReferenceConverterMixin.DB_ID_REFERENCE_TYPE,
        ],
    ):
        self._reference_type = reference_type

    def __get_technical_audit_user(self, technical_user_uuid):
        technical_user = TechnicalUser.objects.get(id=technical_user_uuid)
        core_user = User.objects.filter(t_user=technical_user_uuid).first()
        interactive_user = core_user.i_user
        return interactive_user.id if interactive_user else technical_user.id_for_audit

    def _print_debug_log(self, e):
        import traceback

        debug_log = (
            "FHIR Mapping for Serializer {self} has failed with exception {e}. Traceback: \n"
            + str(traceback.format_stack())
        )
        logger.debug(debug_log)



import logging

from rest_framework.request import Request

from api_fhir_r4.mixins import (
    MultiIdentifierRetrieveManySerializersMixin,
    MultiIdentifierRetrieverMixin,
)
from api_fhir_r4.model_retrievers import (
    UUIDIdentifierModelRetriever,
    CodeIdentifierModelRetriever,
)
from api_fhir_r4.multiserializer import modelViewset
from api_fhir_r4.permissions import (
    FHIRApiPractitionerClaimAdminPermissions,
    FHIRApiPractitionerOfficerPermissions,
)
from api_fhir_r4.serializers import (
    ClaimAdminPractitionerSerializer,
    EnrolmentOfficerPractitionerSerializer,
)
from api_fhir_r4.views.fhir.base import BaseMultiserializerFHIRView
from api_fhir_r4.views.filters import ValidityFromRequestParameterFilter
from core.models.user import ClaimAdmin
from core.models import Officer


logger = logging.getLogger(__name__)


class PractitionerViewSet(
    BaseMultiserializerFHIRView,
    modelViewset.MultiSerializerModelViewSet,
    MultiIdentifierRetrieveManySerializersMixin,
    MultiIdentifierRetrieverMixin,
):
    retrievers = [UUIDIdentifierModelRetriever, CodeIdentifierModelRetriever]

    lookup_field = "identifier"

    @property
    def serializers(self):
        return {
            ClaimAdminPractitionerSerializer: (
                self._ca_queryset(),
                self._ca_serializer_validator,
                (FHIRApiPractitionerClaimAdminPermissions,),
            ),
            EnrolmentOfficerPractitionerSerializer: (
                self._eo_queryset(),
                self._eo_serializer_validator,
                (FHIRApiPractitionerOfficerPermissions,),
            ),
        }

    @classmethod
    def _ca_serializer_validator(cls, context):
        return cls._base_request_validator_dispatcher(
            request=context["request"],
            get_check=lambda x: cls._get_type_from_query(x) in ("ca", None),
            post_check=lambda x: cls._get_type_from_body(x) == "ca",
            put_check=lambda x: cls._get_type_from_body(x) in ("ca", None),
        )

    @classmethod
    def _eo_serializer_validator(cls, context):
        return cls._base_request_validator_dispatcher(
            request=context["request"],
            get_check=lambda x: cls._get_type_from_query(x) in ("eo", None),
            post_check=lambda x: cls._get_type_from_body(x) == "eo",
            put_check=lambda x: cls._get_type_from_body(x) in ("eo", None),
        )

    @classmethod
    def _base_request_validator_dispatcher(
        cls, request: Request, get_check, post_check, put_check
    ):
        if request.method == "GET":
            return get_check(request)
        elif request.method == "POST":
            return post_check(request)
        elif request.method == "PUT":
            return put_check(request)
        return True

    def list(self, request, *args, **kwargs):
        identifier = request.GET.get("code")
        if identifier:
            return self.retrieve(request, *args, **{**kwargs, "identifier": identifier})
        return super().list(request, *args, **kwargs)

    def get_queryset(self):
        return ClaimAdmin.objects

    def _ca_queryset(self):
        queryset = ClaimAdmin.objects.filter(validity_to__isnull=True).all()
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(
            queryset
        )

    def _eo_queryset(self):
        queryset = (
            Officer.objects.filter(validity_to__isnull=True)
            .order_by("validity_from")
            .all()
        )
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(
            queryset
        )

    @classmethod
    def _get_type_from_body(cls, request):
        try:
            # See: http://hl7.org/fhir/R4/practitioner.html
            return request.data["qualification"][0]["code"]["coding"][0]["code"].lower()
        except KeyError:
            logger.exception(
                "Failed to match IMIS practitioner type using request body. It should be accessible under"
                "body.qualification[0].code.coding[0].code"
            )
            return None

    @classmethod
    def _get_type_from_query(cls, request):
        try:
            return request.GET["resourceType"].lower()
        except KeyError:
            return None

    def _raise_no_eligible_serializer(self):
        raise AssertionError(
            "Failed to match serializer eligible for given request. "
            "For POST request JSON should determine eo/ca type in body.type[0].coding[0].code"
        )

    def _raise_multiple_eligible_serializers(self):
        raise AssertionError(
            "Ambiguous request, more than one serializer is eligible for given action. "
        )



from api_fhir_r4.converters import CommunicationRequestConverter
from fhir.resources.R4B.communicationrequest import CommunicationRequest
from api_fhir_r4.tests import CommunicationRequestTestMixin
from api_fhir_r4.tests.mixin import ConvertToFhirTestMixin, ConvertJsonToFhirTestMixin


class CommunicationRequestConverterTestCase(
    CommunicationRequestTestMixin, ConvertToFhirTestMixin, ConvertJsonToFhirTestMixin
):
    converter = CommunicationRequestConverter
    fhir_resource = CommunicationRequest
    json_repr = "test/test_communicationRequest.json"



from rest_framework.viewsets import GenericViewSet

from api_fhir_r4.mixins import MultiIdentifierRetrieverMixin, ListModelMixin
from api_fhir_r4.model_retrievers import UUIDIdentifierModelRetriever
from api_fhir_r4.permissions import FHIRApiCommunicationRequestPermissions
from api_fhir_r4.serializers import CommunicationRequestSerializer
from api_fhir_r4.views.fhir.base import BaseFHIRView
from api_fhir_r4.views.filters import ValidityFromRequestParameterFilter
from claim.models import Claim


class CommunicationRequestViewSet(
    BaseFHIRView, MultiIdentifierRetrieverMixin, ListModelMixin, GenericViewSet
):
    retrievers = [UUIDIdentifierModelRetriever]
    serializer_class = CommunicationRequestSerializer
    permission_classes = (FHIRApiCommunicationRequestPermissions,)

    def get_queryset(self):
        queryset = (
            Claim.get_queryset(None, self.request.user)
            .filter(
                feedback_status__in=[
                    Claim.FEEDBACK_SELECTED,
                    Claim.FEEDBACK_DELIVERED,
                    Claim.FEEDBACK_BYPASSED,
                ]
            )
            .order_by("validity_from")
        )
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(
            queryset
        )



from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.models.imisModelEnums import ContactPointSystem


class HealthFacilityOrganizationTypeMapping:
    LEGAL_FORM_CODE = "D"
    LEGAL_FORM_DISPLAY = "District organization"
    LEGAL_FORM_SYSTEM = "CodeSystem/organization-legal-form"
    LEGAL_FORM_URL = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/organization-legal-form"

    ORGANIZATION_TYPE = "prov"

    EMAIL_CONTACT_POINT_SYSTEM = ContactPointSystem.EMAIL
    PHONE_CONTACT_POINT_SYSTEM = ContactPointSystem.PHONE
    FAX_CONTACT_POINT_SYSTEM = ContactPointSystem.FAX

    ADDRESS_LOCATION_REFERENCE_URL = f"""{
        GeneralConfiguration.get_system_base_url()
    }/StructureDefinition/address-location-reference"""

    CONTRACT_PURPOSE = {
        "code": "PAYOR",
        "system": "http://terminology.hl7.org/CodeSystem/contactentity-type",
    }

    LEVEL_DISPLAY_MAPPING = {"D": "Dispensary", "C": "Health Centre", "H": "Hospital"}

    TYPE_DISPLAY_MAPPING = {"O": "Out-patient", "I": "In-patient", "B": "Both"}

    LEVEL_SYSTEM = (
        f"{GeneralConfiguration.get_system_base_url()}/CodeSystem/organization-hf-level"
    )

    TYPE_SYSTEM = f"{GeneralConfiguration.get_system_base_url()}/CodeSystem/organization-hf-care-type"



from api_fhir_r4.models import Subscription
from api_fhir_r4.tests import GenericTestMixin
from api_fhir_r4.utils import TimeUtils

from fhir.resources.R4B.subscription import Subscription as FHIRSubscription


class SubscriptionTestMixin(GenericTestMixin):
    _TEST_SUB_ENDPOINT = "https://test.com/test"
    _TEST_SUB_HEADERS = "TestKey1=TestValue1&TestKey2=TestValue2"
    _TEST_SUB_IMIS_EXPIRING = "2020-01-01T00:00:00+00:00"

    _TEST_SUB_IMIS_CRITERIA = {"resource": "Patient", "chfid__startswith": "0"}
    _TEST_SUB_IMIS_STATUS = 1
    _TEST_SUB_IMIS_CHANNEL = 1

    _TEST_SUB_FHIR_STATUS = "active"
    _TEST_SUB_FHIR_REASON = "Patient"
    _TEST_SUB_FHIR_CRITERIA = "Patient?chfid__startswith=0"
    _TEST_SUB_FHIR_CHANNEL_TYPE = "rest-hook"

    def create_test_imis_instance(self):
        return Subscription(
            **{
                "status": self._TEST_SUB_IMIS_STATUS,
                "channel": self._TEST_SUB_IMIS_CHANNEL,
                "endpoint": self._TEST_SUB_ENDPOINT,
                "headers": self._TEST_SUB_HEADERS,
                "criteria": self._TEST_SUB_IMIS_CRITERIA,
                "expiring": TimeUtils.str_iso_to_date(self._TEST_SUB_IMIS_EXPIRING),
            }
        )

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(imis_obj.status, self._TEST_SUB_IMIS_STATUS)
        self.assertEqual(imis_obj.channel, self._TEST_SUB_IMIS_CHANNEL)
        self.assertEqual(imis_obj.endpoint, self._TEST_SUB_ENDPOINT)
        self.assertEqual(imis_obj.headers, self._TEST_SUB_HEADERS)
        self.assertEqual(imis_obj.criteria, self._TEST_SUB_IMIS_CRITERIA)
        self.assertEqual(
            imis_obj.expiring, TimeUtils.str_iso_to_date(self._TEST_SUB_IMIS_EXPIRING)
        )

    def create_test_fhir_instance(self):
        return FHIRSubscription.parse_obj(
            {
                "status": self._TEST_SUB_FHIR_STATUS,
                "criteria": self._TEST_SUB_FHIR_CRITERIA,
                "end": self._TEST_SUB_IMIS_EXPIRING,
                "reason": self._TEST_SUB_FHIR_REASON,
                "channel": {
                    "type": self._TEST_SUB_FHIR_CHANNEL_TYPE,
                    "endpoint": self._TEST_SUB_ENDPOINT,
                    "header": [self._TEST_SUB_HEADERS],
                },
            }
        )

    def verify_fhir_instance(self, fhir_obj):
        self.assertEqual(fhir_obj.status, self._TEST_SUB_FHIR_STATUS)
        self.assertEqual(fhir_obj.criteria, self._TEST_SUB_FHIR_CRITERIA)
        self.assertEqual(
            fhir_obj.end, TimeUtils.str_iso_to_date(self._TEST_SUB_IMIS_EXPIRING)
        )
        self.assertEqual(fhir_obj.channel.type, self._TEST_SUB_FHIR_CHANNEL_TYPE)
        self.assertEqual(fhir_obj.channel.endpoint, self._TEST_SUB_ENDPOINT)
        self.assertEqual(fhir_obj.channel.header, [self._TEST_SUB_HEADERS])



from api_fhir_r4.serializers import CodeSystemSerializer

from rest_framework import viewsets
from rest_framework.response import Response
from rest_framework.permissions import IsAuthenticated

from rest_framework.views import APIView
from api_fhir_r4.views import CsrfExemptSessionAuthentication
from policyholder.services import PolicyHolderActivity


class CodeSystemOrganizationPHActivityViewSet(viewsets.ViewSet):
    serializer_class = CodeSystemSerializer
    permission_classes = (IsAuthenticated,)
    authentication_classes = [
        CsrfExemptSessionAuthentication
    ] + APIView.settings.DEFAULT_AUTHENTICATION_CLASSES

    def list(self, request):
        # we don't use typical instance, we only indicate the model and the field to be mapped into CodeSystem
        serializer = CodeSystemSerializer(
            user=request.user,
            instance=None,
            **{
                "data": PolicyHolderActivity(request.user).get_all()["data"],
                "code_field": "code",
                "display_field": "display",
                "id": "organization-ph-activity",
                "name": "OrganizationPHActivityCS",
                "title": "Activity (Organization)",
                "description": "Indicates the activity of the PolicyHolder Organization. "
                "Values defined by openIMIS. Can be extended.",
                "url": self.request.build_absolute_uri(),
            }
        )
        data = serializer.to_representation(obj=None)
        return Response(data)



from api_fhir_r4.converters import ClaimAdminPractitionerConverter

from fhir.resources.R4B.practitioner import Practitioner
from api_fhir_r4.tests import ClaimAdminPractitionerTestMixin
from api_fhir_r4.tests.mixin import (
    ConvertToImisTestMixin,
    ConvertToFhirTestMixin,
    ConvertJsonToFhirTestMixin,
)


class ClaimAdminPractitionerConverterTestCase(
    ClaimAdminPractitionerTestMixin,
    ConvertToImisTestMixin,
    ConvertToFhirTestMixin,
    ConvertJsonToFhirTestMixin,
):
    converter = ClaimAdminPractitionerConverter
    fhir_resource = Practitioner
    json_repr = "test/test_claimAdminPractitioner.json"



from api_fhir_r4.utils import DbManagerUtils
from core.forms import User
from core.services import create_or_update_interactive_user, create_or_update_core_user
from core.utils import set_current_user
from core.test_helpers import create_test_interactive_user, create_admin_role
import json
import os


class LogInMixin:
    _TEST_USER_NAME = "Admin"
    _TEST_USER_PASSWORD = "admin123"
    _TEST_DATA_USER = {
        "username": _TEST_USER_NAME,
        "last_name": _TEST_USER_NAME,
        "password": _TEST_USER_PASSWORD,
        "other_names": _TEST_USER_NAME,
        "user_types": "INTERACTIVE",
        "language": "en",
        "roles": [create_admin_role().id],
    }

    def load_user_data_from_json(self, json_path):
        """
        Purpose: one
        """
        dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        json_representation = json.loads(open(dir_path + json_path).read())
        if "username" in json_representation and "password" in json_representation:
            self._test_username = json_representation["username"]
            self._test_username = json_representation["password"]
        else:
            raise Exception("Json not containing credentials")

    # end def

    def get_or_create_user_api(self):
        user = create_test_interactive_user(username=self._TEST_USER_NAME, password=self._TEST_USER_PASSWORD)
        set_current_user(user)
        return user

    def __create_user_interactive_core(self):
        i_user, i_user_created = create_or_update_interactive_user(
            user_id=None, data=self._TEST_DATA_USER, audit_user_id=999, connected=False
        )
        create_or_update_core_user(
            user_uuid=None, username=self._TEST_DATA_USER["username"], i_user=i_user
        )
        return DbManagerUtils.get_object_or_none(User, username=self._TEST_USER_NAME)



import copy

from .utils import PaymentNoticeSerializerUtils
from invoice.services import PaymentInvoiceService


class PaymentNoticeSerializerCreate(object):
    @classmethod
    def create(cls, validated_data, request):
        user = request.user

        imis_payment_detail = validated_data.pop("imis_payment_detail")
        invoice_status = validated_data.pop("invoice_status")

        payment_invoice = copy.deepcopy(validated_data)
        if "_state" in payment_invoice:
            del payment_invoice["_state"]
        if "_original_state" in payment_invoice:
            del payment_invoice["_original_state"]

        payment_invoice_service = PaymentInvoiceService(user)
        result = payment_invoice_service.create_with_detail(
            payment_invoice, imis_payment_detail
        )
        payment_invoice = PaymentNoticeSerializerUtils.get_result_object(result)
        if invoice_status and result["success"]:
            # change the invoice status based on 'fhir status'
            PaymentNoticeSerializerUtils.change_invoice_status(
                payment_invoice_service, payment_invoice
            )
        return payment_invoice
